# -*- coding: utf-8 -*-
"""
Tools for force limiting.
"""

from types import SimpleNamespace
import numpy as np
import scipy.linalg as la
from scipy.optimize import minimize_scalar
from pyyeti import cb, ode


# FIXME: We need the str/repr formatting used in Numpy < 1.14.
try:
    np.set_printoptions(legacy="1.13")
except TypeError:
    pass


def calcAM(S, freq):
    """
    Calculate apparent mass

    Parameters
    ----------
    S : list/tuple
        Contains: ``[mass, damp, stiff, bdof]`` for structure. These
        are the source mass, damping, and stiffness matrices (see
        :class:`pyyeti.ode.SolveUnc`) and `bdof`, which is described
        below.
    freq : 1d array_like
        Frequency vector (Hz)

    Returns
    -------
    AM : 3d ndarray
        Apparent mass matrix in a 3d array:

        .. code-block:: none

            boundary DOF x Frequencies x boundary DOF
             (response)                   (input)

    Notes
    -----
    The `bdof` input defines boundary DOF in one of two ways as
    follows. Let `N` be total number of DOF in mass, damping, &
    stiffness.

       1.  If `bdof` is a 2d array_like, it is interpreted to be a
           data recovery matrix to the b-set (number b-set =
           ``bdof.shape[0]``). Structure is treated generically (uses
           :class:`pyyeti.ode.SolveUnc` with ``pre_eig=True`` to
           compute apparent mass).
       2.  Otherwise, `bdof` is assumed to be a 1d partition vector
           from full `N` size to b-set and structure is assumed to be
           in Craig-Bampton form (uses :func:`pyyeti.cb.cbtf` to
           compute apparent mass).

    The routine :func:`ntfl` example demonstrates this function.

    See also
    --------
    :func:`ntfl`.
    """
    lf = len(freq)
    m = S[0]
    b = S[1]
    k = S[2]
    bdof = np.atleast_1d(S[3])

    if bdof.ndim == 2:  # bdof is treated as a drm
        r = bdof.shape[0]
        T = bdof
        Frc = np.zeros((r, lf))
        Acc = np.empty((r, lf, r), dtype=complex)
        fs = ode.SolveUnc(m, b, k, pre_eig=True)
        if hasattr(fs.pc, "eig_success") and not fs.pc.eig_success:
            print(
                "Switching from `SolveUnc` to `FreqDirect` because complex"
                " eigensolver failed; see messages above."
            )
            fs = ode.FreqDirect(m, b, k)
        for direc in range(r):
            Frc[direc, :] = 1.0
            sol = fs.fsolve(T.T @ Frc, freq)
            Acc[:, :, direc] = T @ sol.a
            Frc[direc, :] = 0.0
        AM = np.empty((r, lf, r), dtype=complex)
        for j in range(lf):
            AM[:, j, :] = la.inv(Acc[:, j, :])
    else:  # bdof treated as a partition vector for CB model
        r = len(bdof)
        acce = np.eye(r)
        # Perform Baseshake
        # cbtf = craig bampton transfer function; this will genenerate
        # the corresponding interface force required to meet imposed
        # acceleration
        AM = np.empty((r, lf, r), dtype=complex)
        save = {}
        for direc in range(r):
            tf = cb.cbtf(m, b, k, acce[direc, :], freq, bdof, save)
            AM[:, :, direc] = tf.frc
    return AM


def ntfl(Source, Load, As, freq):
    r"""
    Norton Thevenin Force Limit

    Parameters
    ----------
    Source : list/tuple or 3d ndarray
        Can be either:

           1. list/tuple of ``[mass, damp, stiff, bdof]`` for source
              (eg, launch vehicle). These are the source mass,
              damping, and stiffness matrices (see
              :class:`pyyeti.ode.SolveUnc`) and `bdof`, which is
              described below.
           2. SAM, a 3d ndarray of source apparent mass (from a
              previous run). See description of outputs.

    Load :  list/tuple or 3d ndarray
        Same format as `Source` except for the "load" (eg spacecraft)
    As : 2d array_like
        Free acceleration of the source (interface acceleration
        without the Load attached).
    freq : 1d array_like
        Frequency vector in Hz for `Source`, `Load`, `As` and all
        return values.

    Returns
    -------
    A SimpleNamespace with the members:

    R : 2d ndarray
        Norton Thevenin normalized response ratio; complex,
        # bdof x len(freq). `R` is independent of `As`. Each row of
        `R` is the ratio of loaded-response over free-response
        assuming a unit input in that direction (with the other
        directions fixed).
    A : 2d ndarray
        Coupled system interface acceleration, complex, # bdof x
        len(freq).
    F : 2d ndarray
        Coupled system interface force, complex, # bdof x len(freq)
    LAM, SAM, TAM : 3d ndarrays
        Load, Source and Total apparent mass matrices:

        .. code-block:: none

            boundary DOF x Frequencies x boundary DOF
             (response)                   (input)

        ``TAM = SAM + LAM``

    freq : 1d array_like
        Copy of the input `freq`.

    Notes
    -----
    The outputs are computed as follows:

    .. math::

        \begin{aligned}
        R(f) &= diag(T_{AM}(f)^{-1} \cdot L_{AM}(f)) \\
        A(f) &= T_{AM}(f)^{-1} \cdot L_{AM}(f) \cdot A_s(f) \\
        F(f) &= S_{AM}(f) \cdot T_{AM}(f)^{-1} \cdot L_{AM}(f) \cdot A_s(f)
        \end{aligned}

    The `bdof` input defines boundary DOF in one of two ways as
    follows. Let `N` be total number of DOF in mass, damping, &
    stiffness.

       1.  If `bdof` is a 2d array_like, it is interpreted to be a
           data recovery matrix to the b-set (number b-set =
           ``bdof.shape[0]``). Structure is treated generically (uses
           :class:`pyyeti.ode.SolveUnc` with ``pre_eig=True`` to
           compute apparent mass).
       2.  Otherwise, `bdof` is assumed to be a 1d partition vector
           from full `N` size to b-set and structure is assumed to be
           in Craig-Bampton form (uses :func:`pyyeti.cb.cbtf` to
           compute apparent mass).

    Note that the Source and Load `bdof` must define the same number
    of boundary DOF and both sets of boundary DOF must be in the same
    coordinate system.

    Tips:

    - If using a free-free model, include residual vectors defined
      relative to the boundary DOF.
    - The free acceleration of the source is the complex response of
      the boundary DOF without the load attached. Frequency envelopes
      of flight data, or vibration specifications derived from
      envelopes, are not accurate ways to come up with the free
      acceleration.

    Notional example::

        from pyyeti import frclim
        from pyyeti.nastran op2, n2p, op4
        import pickle

        # Load free acceleration of LV
        dct = pickle.load('ifresults_free.p')
        As = dct['As']
        freq = dct['freq']

        # Load source free-free model, with residual vectors included
        nas = op2.rdnas2cam('nas2cam')
        m1 = None
        zeta = 0.01
        k1 = nas['lambda'][0]
        k1[:nas['nrb']] = 0.
        b1 = 2*zeta*np.sqrt(k1)

        # Transformation to i/f node:
        T, dof = n2p.formdrm(nas, seup=0, sedn=0, dof=888888)

        # Load S/C mass and stiffness:
        mk = op4.read('mk.op4')
        kgen = mk['kgen']
        mgen = mk['mgen']
        kgen[:6, :6] = 0.
        zeta = 0.01
        bgen = np.diag(2*zeta*np.sqrt(np.diag(kgen)))

        # Norton Thevenin force limit function:
        r = frclim.ntfl([m1, b1, k1, T], [mgen, bgen, kgen,
                        np.arange(6)], As, freq)

    See also
    --------
    :func:`sefl`, :func:`ctdfs`, :func:`stdfs`.

    Examples
    --------
    This example sets up a simple mass-spring system to demonstrate
    that the Norton-Thevenin equations can be exact.

    Steps:

      1. setup a coupled system of a SOURCE and a LOAD
      2. solve for interface acceleration and force from coupled system
         (frequency domain)
      3. calculate free acceleration from SOURCE alone and setup LOAD
      4. use :func:`ntfl` to couple the system
      5. plot interface acceleration and force to show :func:`ntfl` can
         be an exact coupling method
      6. plot apparent masses
      7. plot free acceleration and coupled acceleration
      8. plot normalized response ratio (can be thought of as force
         limiting factor)

    1. Setup system:

    .. code-block:: none

                |--> x1       |--> x2        |--> x3        |--> x4
                |             |              |              |
             |----|    k1   |----|    k2   |----|    k3   |----|
         Fe  |    |--\/\/\--|    |--\/\/\--|    |--\/\/\--|    |
        ====>| 10 |         | 30 |         |  3 |         |  2 |
             |    |---| |---|    |---| |---|    |---| |---|    |
             |----|    c1   |----|    c2   |----|    c3   |----|

             |<--- SOURCE --->||<------------ LOAD ----------->|

    Define parameters:

    .. plot::
        :context: close-figs

        >>> import numpy as np
        >>> from pyyeti import frclim, ode
        >>> freq = np.arange(0.0, 25.1, 0.1)
        >>> M1 = 10.
        >>> M2 = 30.
        >>> M3 = 3.
        >>> M4 = 2.
        >>> c1 = 15.
        >>> c2 = 15.
        >>> c3 = 15.
        >>> k1 = 45000.
        >>> k2 = 25000.
        >>> k3 = 10000.

        2. Solve coupled system:

        >>> MASS = np.array([[M1, 0, 0, 0],
        ...                  [0, M2, 0, 0],
        ...                  [0, 0, M3, 0],
        ...                  [0, 0, 0, M4]])
        >>> DAMP = np.array([[c1, -c1, 0, 0],
        ...                  [-c1, c1+c2, -c2, 0],
        ...                  [0, -c2, c2+c3, -c3],
        ...                  [0, 0, -c3, c3]])
        >>> STIF = np.array([[k1, -k1, 0, 0],
        ...                  [-k1, k1+k2, -k2, 0],
        ...                  [0, -k2, k2+k3, -k3],
        ...                  [0, 0, -k3, k3]])
        >>> F = np.vstack((np.ones((1, len(freq))),
        ...                np.zeros((3, len(freq)))))
        >>> fs = ode.SolveUnc(MASS, DAMP, STIF, pre_eig=True)
        >>> fullsol = fs.fsolve(F, freq)
        >>> A_coupled = fullsol.a[1]
        >>> F_coupled = (M2/2*A_coupled -
        ...              k2*(fullsol.d[2] - fullsol.d[1]) -
        ...              c2*(fullsol.v[2] - fullsol.v[1]))

        3. Solve for free acceleration; SOURCE setup: [m, b, k, bdof]:

        >>> ms = np.array([[M1, 0], [0, M2/2]])
        >>> cs = np.array([[c1, -c1], [-c1, c1]])
        >>> ks = np.array([[k1, -k1], [-k1, k1]])
        >>> source = [ms, cs, ks, [[0, 1]]]
        >>> fs_source = ode.SolveUnc(ms, cs, ks, pre_eig=True)
        >>> sourcesol = fs_source.fsolve(F[:2], freq)
        >>> As = sourcesol.a[1:2]   # free acceleration

        LOAD setup: [m, b, k, bdof]:

        >>> ml = np.array([[M2/2, 0, 0], [0, M3, 0], [0, 0, M4]])
        >>> cl = np.array([[c2, -c2, 0], [-c2, c2+c3, -c3], [0, -c3, c3]])
        >>> kl = np.array([[k2, -k2, 0], [-k2, k2+k3, -k3], [0, -k3, k3]])
        >>> load = [ml, cl, kl, [[1, 0, 0]]]

        4. Use NT to couple equations. First value (rigid-body motion)
        should equal ``Source Mass / Total Mass = 25/45 = 0.55555...``
        Results should match the coupled method.

        >>> r = frclim.ntfl(source, load, As, freq)
        >>> abs(r.R[0, 0])   # doctest: +ELLIPSIS
        0.55555...
        >>> np.allclose(A_coupled, r.A)
        True
        >>> np.allclose(F_coupled, r.F)
        True

        Calculate the total apparent mass directly using :func:`calcAM`.
        This should match the ``r.TAM`` value.

        >>> TAM = frclim.calcAM((MASS, DAMP, STIF, [[0, 1, 0, 0]]), freq)
        >>> np.allclose(TAM, r.TAM)
        True
        >>> np.allclose(TAM, r.SAM+r.LAM)
        True

        5. Plot comparisons:

        >>> import matplotlib.pyplot as plt
        >>> _ = plt.figure('Example')
        >>> plt.clf()
        >>> _ = plt.subplot(211)
        >>> _ = plt.semilogy(freq, abs(A_coupled),
        ...                  label='Coupled')
        >>> _ = plt.semilogy(freq, abs(r.A).T, '--',
        ...                  label='NT')
        >>> _ = plt.title('Interface Acceleration')
        >>> _ = plt.xlabel('Frequency (Hz)')
        >>> _ = plt.legend(loc='best')
        >>> _ = plt.subplot(212)
        >>> _ = plt.semilogy(freq, abs(F_coupled),
        ...                  label='Coupled')
        >>> _ = plt.semilogy(freq, abs(r.F).T, '--',
        ...                  label='NT')
        >>> _ = plt.title('Interface Force')
        >>> _ = plt.xlabel('Frequency (Hz)')
        >>> _ = plt.legend(loc='best')
        >>> plt.tight_layout()

    .. plot::
        :context: close-figs

        6. Plot apparent masses:

        >>> _ = plt.figure('Example 2')
        >>> plt.clf()
        >>> _ = plt.semilogy(freq, abs(r.TAM[0, :, 0]),
        ...                  label='Total App. Mass')
        >>> _ = plt.semilogy(freq, abs(r.SAM[0, :, 0]),
        ...                  label='Source App. Mass')
        >>> _ = plt.semilogy(freq, abs(r.LAM[0, :, 0]),
        ...                  label='Load App. Mass')
        >>> _ = plt.title('Apparent Mass Comparison')
        >>> _ = plt.xlabel('Frequency (Hz)')
        >>> _ = plt.legend(loc='best')

    .. plot::
        :context: close-figs

        7. Plot accelerations and
        8. Plot force limit factor:

        >>> _ = plt.figure('Example 3')
        >>> plt.clf()
        >>> _ = plt.subplot(211)
        >>> _ = plt.semilogy(freq, abs(As).T,
        ...                  label='Free Acce')
        >>> _ = plt.semilogy(freq, abs(r.A).T,
        ...                  label='Coupled Acce')
        >>> _ = plt.title('Interface Acceleration')
        >>> _ = plt.xlabel('Frequency (Hz)')
        >>> _ = plt.legend(loc='best')
        >>> _ = plt.subplot(212)
        >>> _ = plt.semilogy(freq, abs(r.R).T)
        >>> _ = plt.title('NT Response Ratio: '
        ...               'R = Coupled Acce / Free Acce')
        >>> _ = plt.xlabel('Frequency (Hz)')
        >>> plt.tight_layout()
    """
    # Calculate apparent masses:
    if isinstance(Source, (list, tuple)):
        SAM = calcAM(Source, freq)
    else:
        SAM = Source

    if isinstance(Load, (list, tuple)):
        LAM = calcAM(Load, freq)
    else:
        LAM = Load

    As = np.atleast_2d(As)
    if not len(freq) == As.shape[1] == SAM.shape[1] == LAM.shape[1]:
        raise ValueError(
            "incompatible sizes: ensure that `Source`, "
            "`Load`, and `As` all use the same frequency "
            "vector `freq`"
        )

    TAM = SAM + LAM

    # Application of Norton-Thevenin equations
    r, c, _ = SAM.shape
    R = np.empty((r, c), dtype=complex)
    A = np.empty((r, c), dtype=complex)
    F = np.empty((r, c), dtype=complex)
    for j in range(c):
        Ms = SAM[:, j, :]
        Ml = LAM[:, j, :]
        Mr = la.solve(Ms + Ml, Ms)
        R[:, j] = np.diag(Mr)
        A[:, j] = Mr @ As[:, j]
        F[:, j] = Ml @ A[:, j]
    return SimpleNamespace(R=R, F=F, A=A, LAM=LAM, SAM=SAM, TAM=TAM, freq=freq)


def sefl(c, f, f0):
    """
    Semi-empirical normalized force limit.

    Parameters
    ----------
    c : scalar
        Constant based on experience, typically around 1.5
    f : scalar
        Frequency of interest, typically lower end of band.
    f0 : scalar
        Fundamental frequency in direction of interest.

    Returns
    -------
    nfl : scalar
        The normalized force limit:

        .. code-block:: none

            nfl = c                 f <= f0
            nfl = c / (f/f0)        f > f0

    Notes
    -----
    See reference [#fl1]_ for more information on force limiting.

    References
    ----------
    .. [#fl1] Scharton, T.D. (1997). 'Force Limited Vibration Testing
        Monograph'. NASA. Reference Publication RP-1403, JPL.

    See also
    --------
    :func:`ntfl`, :func:`ctdfs`, :func:`stdfs`.

    Examples
    --------
    Compute force limit for a s/c attached to a launch vehicle, where
    the interface acceleration specification level is 1.75 g, and:

     - frequency of interest starts at 75 Hz
     - fundamental axial mode of s/c is 40 Hz

    >>> from pyyeti import frclim
    >>> m = 6961    # s/c mass
    >>> faf = 40    # fundamental axial frequency of s/c
    >>> spec = 1.75
    >>> m*spec*frclim.sefl(1.5, 75, faf)
    9745.4
    """
    if f <= f0:
        return c
    return f0 * c / f


def stdfs(mr, Q):
    r"""
    Compute the normalized force limit for simple 2-DOF system.

    Parameters
    ----------
    mr : scalar
        Mass ratio m2/m1; 0.0001 to 10 is reasonable.
    Q : scalar to 2 element array_like
        Dynamic amplification factor, 1/2/zeta. If a scalar, the same
        value is used for both dampers. If a 2 element vector, it is
        [Q1, Q2] (see figure below).

    Returns
    -------
    nfl : scalar
        The normalized force limit for m2:

        .. code-block:: none

            nfl = force_limit / (m2 * max(a1))

    Notes
    -----
    The simple 2-DOF system:

    .. code-block:: none

            |---> a0
            |                     |---> a1           |---> a2
        |---------|               |                  |
        |         |     k1     |-----|     k2     |-----|
        |  rigid  |----\/\/\---|     |----\/\/\---|     |
        |  base   |            | m1  |            | m2  |
        |  (M)    |----| |-----|     |----| |-----|     |
        |         |     c1     |-----|     c2     |-----|
        |---------|           (source)            (load)

    Analysis is done in the frequency domain. Methodology:

      1. Define stiffnesses such that ``k1/m1 = k2/m2`` (this is worst
         case, or very near)
      2. Excite M, bounding frequency range of interest
      3. Recover maximum interface acceleration: ``A`` (accel of m1)
      4. Recover maximum interface force on m2: ``F``
      5. Compute normalized force limit: ``F / (A * m2)``

    Note: higher masses result in higher force limits.

    For multi-dof systems, m1 and m2 can be defined as the modal mass
    in the frequency range of interest.  Or, for more conservatism, m1
    and m2 can be defined as the modal mass plus any residual mass.

    The modal masses are defined relative to the interface point. See
    references [#fl2]_ and [#fl3]_ for more information.

    References
    ----------
    .. [#fl2] Scharton, T.D. (1997). 'Force Limited Vibration Testing
        Monograph'. NASA. Reference Publication RP-1403, JPL.
    .. [#fl3] Y. Soucy, A. Cote, 'Reduction of Overtesting during
        Vibration Tests of Space Hardware', Can. Aeronautics and Space
        J., Vol. 48, No. 1, pp. 77-86, 2002.

    See also
    --------
    :func:`ntfl`, :func:`ctdfs`, :func:`sefl`.

    Examples
    --------
    Compute force limit for a s/c attached to a launch vehicle, where
    the interface acceleration specification level is 1.75 g and Q is
    assumed to be 10:

    >>> from pyyeti import frclim
    >>> m1 = 710     # modal mass + residual mass of lv
    >>> m2 = 3060    # modal mass + residual mass of s/c
    >>> Q = 10
    >>> spec = 1.75
    >>> frclim.stdfs(m2/m1, Q) * m2 * 1.75   # doctest: +ELLIPSIS
    6393.1622...

    Generate some curves showing the normalized force limit vs the
    mass ration for a few different damping values:

    .. plot::
        :context: close-figs

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from pyyeti import frclim
        >>> mr = np.geomspace(0.00001, 100, 100)
        >>> for Q in (10, 25, 50):
        ...     nfl = [frclim.stdfs(i, Q) for i in mr]
        ...     _ = plt.loglog(mr, nfl, label=f'Q={Q}')
        >>> _ = plt.legend()
        >>> _ = plt.xlabel('Mass ratio (Load/Source)')
        >>> _ = plt.ylabel('Normalized Force Limit')
    """
    m2 = 0.1
    M = 1e6
    w1 = 1.0
    w2 = 1.0
    Q = np.atleast_1d(Q)
    if len(Q) == 1:
        zeta1 = zeta2 = 1 / 2 / Q[0]
    else:
        zeta1 = 1 / 2 / Q[0]
        zeta2 = 1 / 2 / Q[1]
    m1 = m2 / mr
    c1 = 2 * zeta1 * w1 * m1
    c2 = 2 * zeta2 * w2 * m2
    k1 = w1 ** 2 * m1
    k2 = w2 ** 2 * m2

    # setup system equations of motion:
    mass = np.array([[M, 0, 0], [0, m1, 0], [0, 0, m2]])
    damp = np.array([[c1, -c1, 0], [-c1, c1 + c2, -c2], [0, -c2, c2]])
    stif = np.array([[k1, -k1, 0], [-k1, k1 + k2, -k2], [0, -k2, k2]])
    w = la.eigh(stif, mass, eigvals_only=True)

    # lam = np.sqrt(abs(w))/w2
    # freq = w2*np.arange(.2, 5, 0.001)/2/np.pi
    # F = np.zeros((3, len(freq)))
    lam = np.sqrt(abs(w))
    freq = lam[1] / 2 / np.pi
    F = np.zeros((3, 1))
    F[0] = 1e6
    fs = ode.FreqDirect(mass, damp, stif)
    sol = fs.fsolve(F, freq)

    spec_level = np.max(abs(sol.a[1]))
    ifforce = np.max(abs(m2 * sol.a[2]))
    return ifforce / spec_level / m2


def _ctdfs_old(mmr1, mmr2, rmr, Q, wr=(1 / np.sqrt(2), np.sqrt(2))):
    r"""
    Compute the normalized force limit for complex 2-DOF system.

    Parameters
    ----------
    mmr1 : scalar
        Modal to residual mass ratio for source; 0.0001 to 10 is
        reasonable ``m1/M1``
    mmr2 : scalar
        Modal to residual mass ratio for load; 0.0001 to 10 is
        reasonable ``m2/M2``
    rmr : scalar
        Residual mass ratio of source over load; 0.0001 to 10 is
        reasonable ``M2/M1``
    Q : scalar to 2 element array_like
        Dynamic amplification factor, 1/2/zeta. If a scalar, the same
        value is used for both dampers. If a 2 element vector, it is
        [Q1, Q2] (see figure below).
    wr : 2 element array_like

        Two element tuning range for frequency of LOAD. `wr` is a
        ratio of the LOAD frequency to the SOURCE::

            wr = [ w2_min w2_max ] / w1

        Scharton used [1/sqrt(2), sqrt(2)] in [#fl4]_.

    Returns
    -------
    nfl : scalar
        Normalized force limit for M2::

            nfl = force_limit / (M2 * max(a2))

    nw2 : scalar
        Normalized tuned frequency of LOAD::

            nw2 = w2_tuned / w1

    Notes
    -----
    This routine computes the normalize force limit for M2 in the
    complex 2-DOF system (2 flexible body modes)::


               |--> a1         |--> a2        |--> a3         |--> a4
               |               |              |               |
             |----|    k1    |----|    F    |----|    k2    |----|
         F1  |    |---\/\/\--|    |<------->|    |---\/\/\--|    |
        ---->| m1 |          | M1 |         | M2 |          | m2 |
             |    |---| |----|    |  a2=a3  |    |---| |----|    |
             |----|    c1    |----|<------->|----|    c2    |----|
              modal         residual       residual          modal
                    S O U R C E                     L O A D

    Analysis is done in the frequency domain.  Methodology:

      1. Set ``w1 = sqrt(k1/m1) = 1`` and ``M1 = 1``
      2. Tune ``w2 = sqrt(k2/m2)`` such that a worst case force limit
         is achieved within pre-defined frequency limits according to
         input `wr`. Frequency range is limited because modal and
         residual masses as input are only valid in a limited
         frequency range.  For each w2:

         a. Compute natural frequencies and solve equations of motion
            at the two flexible frequencies. (Note: this is not
            guaranteed to be the worst-case frequencies: almost 10%
            "errors" have been seen for ``Q = 5`` systems. Lower
            damping is much closer.)
         b. Recover maximum interface accel: ``A`` (accel of M2)
         c. Recover maximum interface force on M2: ``F``
         d. Compute `nfl`: ``F / (A * M2)``

      3. Keep the maximum `nfl` from 2.

    The modal masses are defined relative to the interface point. See
    references [#fl4]_ and [#fl5]_ for more information.

    References
    ----------
    .. [#fl4] Scharton, T.D. (1997). 'Force Limited Vibration Testing
        Monograph'. NASA. Reference Publication RP-1403, JPL.
    .. [#fl5] Y. Soucy, A. Cote, 'Reduction of Overtesting during
        Vibration Tests of Space Hardware', Can. Aeronautics and Space
        J., Vol. 48, No. 1, pp. 77-86, 2002.

    See also
    --------
    :func:`ntfl`, :func:`stdfs`, :func:`sefl`.

    Examples
    --------
    Compute force limit for a s/c attached to a launch vehicle, where
    the interface acceleration specification level is 1.75 g and Q is
    assumed to be 10. Compare against the :func:`stdfs` and
    :func:`sefl` methods:

    >>> from pyyeti import frclim
    >>> m1 = 30     # lv modal mass 75-90 Hz
    >>> M1 = 622    # lv residual mass above 90 Hz
    >>> m2 = 972    # sc modal mass 75-90 Hz
    >>> M2 = 954    # sc residual mass above 90 Hz
    >>> msc = 6961  # total sc mass
    >>> faf = 40    # fundamental axial frequency of s/c
    >>> Q = 10
    >>> spec = 1.75
    >>> (frclim._ctdfs_old(m1/M1, m2/M2, M2/M1, Q)[0] *
    ...    M2 * spec)                             # doctest: +ELLIPSIS
    8686.1...
    >>> (frclim.stdfs((m2+M2)/(m1+M1), Q) *
    ...    (m2+M2) * spec)                        # doctest: +ELLIPSIS
    4268.2...
    >>> frclim.sefl(1.5, 75, faf) * msc * spec
    9745.4
    """
    M1 = 1.0
    w1 = 1.0
    Q = np.atleast_1d(Q)
    if len(Q) == 1:
        zeta1 = zeta2 = 1 / 2 / Q[0]
    else:
        zeta1 = 1 / 2 / Q[0]
        zeta2 = 1 / 2 / Q[1]

    M2 = M1 * rmr
    m1 = M1 * mmr1
    m2 = M2 * mmr2

    if m2 == 0:
        return 1, 1
    if m1 == 0:
        m1 = 1e-5

    c1 = 2 * zeta1 * w1 * m1
    k1 = w1 ** 2 * m1
    fl = wr[0]  # low bound
    fh = wr[1]  # high bound

    wrange = np.logspace(np.log10(fl), np.log10(fh), 30)
    step = wrange[-1] - wrange[-2]
    maxstep = 0.0001
    tol = 1e-6
    err = 1
    J = 0
    last = 0
    F = np.zeros((3, 2))
    F[0] = 1.0

    mass = np.array([[m1, 0, 0], [0, M1 + M2, 0], [0, 0, m2]])
    while True:
        J += 1
        pknfl = np.zeros_like(wrange)
        for j, w2 in enumerate(wrange * w1):  # tuning loop
            c2 = 2 * zeta2 * w2 * m2
            k2 = w2 ** 2 * m2

            # solve equations of motion at eigenvalues
            damp = np.array([[c1, -c1, 0], [-c1, c1 + c2, -c2], [0, -c2, c2]])
            stif = np.array([[k1, -k1, 0], [-k1, k1 + k2, -k2], [0, -k2, k2]])
            w = la.eigh(stif, mass, eigvals_only=True)
            lam = np.sqrt(abs(w))
            fq2 = lam[1] / 2 / np.pi
            fq3 = lam[2] / 2 / np.pi
            freq = np.array([fq2, fq3])

            # method 1: solve system only at natural frequencies
            fs = ode.FreqDirect(mass, damp, stif)
            sol = fs.fsolve(F, freq)
            d4 = sol.d[2]
            d2 = sol.d[1]
            a2 = sol.a[1]
            v2 = sol.v[1]
            v4 = sol.v[2]
            ifforce = abs(k2 * (d4 - d2) - M2 * a2 + c2 * (v4 - v2))
            ifaccel = abs(a2)
            pknfl[j] = max(ifforce) / max(ifaccel) / M2

        i = np.argmax(pknfl)
        nfl = pknfl[i]
        if J > 1 and step < maxstep:
            err = abs(last / nfl - 1)
            if err < tol:
                nw2 = wrange[i]
                break
        last = nfl
        fl_new = fl if i == 0 else wrange[i - 1]
        fh_new = fh if i == len(wrange) - 1 else wrange[i + 1]
        wrange = np.logspace(np.log10(fl_new), np.log10(fh_new), 30)
        step = wrange[-1] - wrange[-2]
    return nfl, nw2


def ctdfs(mmr1, mmr2, rmr, Q, wr=(1 / np.sqrt(2), np.sqrt(2))):
    r"""
    Compute the normalized force limit for complex 2-DOF system.

    Parameters
    ----------
    mmr1 : scalar
        Modal to residual mass ratio for source; 0.0001 to 10 is
        reasonable ``m1/M1``
    mmr2 : scalar
        Modal to residual mass ratio for load; 0.0001 to 10 is
        reasonable ``m2/M2``
    rmr : scalar
        Residual mass ratio of source over load; 0.0001 to 10 is
        reasonable ``M2/M1``
    Q : scalar to 2 element array_like
        Dynamic amplification factor, 1/2/zeta. If a scalar, the same
        value is used for both dampers. If a 2 element vector, it is
        [Q1, Q2] (see figure below).
    wr : 2 element array_like

        Two element tuning range for frequency of LOAD. `wr` is a
        ratio of the LOAD frequency to the SOURCE::

            wr = [ w2_min w2_max ] / w1

        Scharton used [1/sqrt(2), sqrt(2)] in [#fl4]_.

    Returns
    -------
    nfl : scalar
        Normalized force limit for M2::

            nfl = force_limit / (M2 * max(a2))

    nw2 : scalar
        Normalized tuned frequency of LOAD::

            nw2 = w2_tuned / w1

    Notes
    -----
    This routine computes the normalize force limit for M2 in the
    complex 2-DOF system (2 flexible body modes)::


               |--> a1         |--> a2        |--> a3         |--> a4
               |               |              |               |
             |----|    k1    |----|    F    |----|    k2    |----|
         F1  |    |---\/\/\--|    |<------->|    |---\/\/\--|    |
        ---->| m1 |          | M1 |         | M2 |          | m2 |
             |    |---| |----|    |  a2=a3  |    |---| |----|    |
             |----|    c1    |----|<------->|----|    c2    |----|
              modal         residual       residual          modal
                    S O U R C E                     L O A D

    Analysis is done in the frequency domain.  Methodology:

      1. Set ``w1 = sqrt(k1/m1) = 1`` and ``M1 = 1``
      2. Tune ``w2 = sqrt(k2/m2)`` such that a worst case force limit
         is achieved within pre-defined frequency limits according to
         input `wr`. Frequency range is limited because modal and
         residual masses as input are only valid in a limited
         frequency range.  For each w2:

         a. Compute natural frequencies and solve equations of motion
            at the two flexible frequencies. (Note: this is not
            guaranteed to be the worst-case frequencies: almost 10%
            "errors" have been seen for ``Q = 5`` systems. Lower
            damping is much closer.)
         b. Recover maximum interface accel: ``A`` (accel of M2)
         c. Recover maximum interface force on M2: ``F``
         d. Compute `nfl`: ``F / (A * M2)``

      3. Keep the maximum `nfl` from 2.

    The optimization is carried out by
    :func:`scipy.optimize.minimize_scalar`.

    The modal masses are defined relative to the interface point. See
    references [#fl4]_ and [#fl5]_ for more information.

    References
    ----------
    .. [#fl4] Scharton, T.D. (1997). 'Force Limited Vibration Testing
        Monograph'. NASA. Reference Publication RP-1403, JPL.
    .. [#fl5] Y. Soucy, A. Cote, 'Reduction of Overtesting during
        Vibration Tests of Space Hardware', Can. Aeronautics and Space
        J., Vol. 48, No. 1, pp. 77-86, 2002.

    See also
    --------
    :func:`ntfl`, :func:`stdfs`, :func:`sefl`.

    Examples
    --------
    Compute force limit for a s/c attached to a launch vehicle, where
    the interface acceleration specification level is 1.75 g and Q is
    assumed to be 10. Compare against the :func:`stdfs` and
    :func:`sefl` methods:

    >>> from pyyeti import frclim
    >>> m1 = 30     # lv modal mass 75-90 Hz
    >>> M1 = 622    # lv residual mass above 90 Hz
    >>> m2 = 972    # sc modal mass 75-90 Hz
    >>> M2 = 954    # sc residual mass above 90 Hz
    >>> msc = 6961  # total sc mass
    >>> faf = 40    # fundamental axial frequency of s/c
    >>> Q = 10
    >>> spec = 1.75
    >>> (frclim.ctdfs(m1/M1, m2/M2, M2/M1, Q)[0] *
    ...    M2 * spec)                             # doctest: +ELLIPSIS
    8686.1...
    >>> (frclim.stdfs((m2+M2)/(m1+M1), Q) *
    ...    (m2+M2) * spec)                        # doctest: +ELLIPSIS
    4268.2...
    >>> frclim.sefl(1.5, 75, faf) * msc * spec
    9745.4
    """
    M1 = 1.0
    w1 = 1.0
    Q = np.atleast_1d(Q)
    if len(Q) == 1:
        zeta1 = zeta2 = 1 / 2 / Q[0]
    else:
        zeta1 = 1 / 2 / Q[0]
        zeta2 = 1 / 2 / Q[1]

    M2 = M1 * rmr
    m1 = M1 * mmr1
    m2 = M2 * mmr2

    if m2 == 0:
        return 1, 1
    if m1 == 0:
        m1 = 1e-5

    c1 = 2 * zeta1 * w1 * m1
    k1 = w1 ** 2 * m1

    def get_neg_pknfl(nw2):
        """
        Computes the negative (for minimization) of peak normalized
        force limit
        """
        w2 = nw2 * w1
        c2 = 2 * zeta2 * w2 * m2
        k2 = w2 ** 2 * m2

        # solve equations of motion at eigenvalues
        damp = np.array([[c1, -c1, 0], [-c1, c1 + c2, -c2], [0, -c2, c2]])
        stif = np.array([[k1, -k1, 0], [-k1, k1 + k2, -k2], [0, -k2, k2]])
        w = la.eigh(stif, mass, eigvals_only=True)
        lam = np.sqrt(abs(w))
        fq2 = lam[1] / 2 / np.pi
        fq3 = lam[2] / 2 / np.pi
        freq = np.array([fq2, fq3])

        # method 1: solve system only at natural frequencies
        fs = ode.FreqDirect(mass, damp, stif)
        sol = fs.fsolve(F, freq)
        d4 = sol.d[2]
        d2 = sol.d[1]
        a2 = sol.a[1]
        v2 = sol.v[1]
        v4 = sol.v[2]
        ifforce = abs(k2 * (d4 - d2) - M2 * a2 + c2 * (v4 - v2))
        ifaccel = abs(a2)
        return -max(ifforce) / max(ifaccel) / M2

    F = np.zeros((3, 2))
    F[0] = 1.0
    mass = np.array([[m1, 0, 0], [0, M1 + M2, 0], [0, 0, m2]])
    res = minimize_scalar(get_neg_pknfl, bracket=wr)
    if "message" in res:
        raise RuntimeError(
            f"routine :func:`scipy.optimize.minimize_scalar` failed: {res.message}"
        )
    nw2 = res.x
    nfl = -res.fun
    return nfl, nw2



# -*- coding: utf-8 -*-
"""
Time and frequency domain ODE solvers for matrix equations. Adapted
and enhanced from the Yeti versions (which were adapted and enhanced
from the original CAM versions). Note that some features depend on the
equations being in modal space (particularly important where there are
distinctions between the rigid-body modes and the elastic modes).

.. note::
    Some features of this module are demonstrated in the pyYeti
    :ref:`tutorial`: :doc:`/tutorials/ode`. There is also a link to
    the source Jupyter notebook at the top of the tutorial.
"""

from ._utilities import *
from .freqdirect import FreqDirect
from .frf_mode_participation import getmodepart, modeselect
from .solvecdf import SolveCDF
from .solveexp1 import SolveExp1
from .solveexp2 import SolveExp2
from .solvenewmark import SolveNewmark
from .solveunc import SolveUnc



# -*- coding: utf-8 -*-
"""
Low level tool for tables of results. Typically, this is called
via: :func:`cla.DR_Results.rpttab`.
"""
import copy
import numpy as np
import xlsxwriter
from pyyeti import ytools, writer
from ._utilities import nan_absmax, _get_rpt_headers, _get_numform


__all__ = ["rpttab1"]


# FIXME: We need the str/repr formatting used in Numpy < 1.14.
try:
    np.set_printoptions(legacy="1.13")
except TypeError:
    pass


def _event_count(mat, j, filter_=1e-6):
    n = mat.shape[1]
    count = np.zeros(n, np.int64)
    peaks = mat[np.arange(mat.shape[0]), j]
    pv = (abs(peaks) > filter_).nonzero()[0]
    if pv.size > 0:
        j = j[pv]
        for k in range(n):
            count[k] = np.sum(j == k)
        return count, count / len(pv) * 100
    return count, count


def _add_zero_case(mat, cases):
    pv = (mat == 0).all(axis=1).nonzero()[0]
    if cases is None:
        new_cases = ["N/A"] * mat.shape[0]
    else:
        new_cases = copy.copy(cases)
    for i in pv:
        new_cases[i] = "zero row"
    return new_cases


def _get_absmax(res):
    amx, pv = nan_absmax(res.mx, res.mn)
    aext, pv = nan_absmax(res.ext[:, 0], res.ext[:, 1])
    if res.maxcase is not None:
        amxcase = res.maxcase[:]
        for j in pv.nonzero()[0]:
            amxcase[j] = res.mincase[j]
    else:
        amxcase = None
    return amx, amxcase, aext


def _add_max_plus_min(ec):
    count = ec["Maximum"][0] + ec["Minimum"][0]
    sumcount = count.sum()
    if sumcount > 0:
        countperc = 100 * count / sumcount
    else:
        countperc = count
    ec["Max+Min"] = {0: count, 1: countperc}


def _rowlbls_table(lbl, ec, j):
    rowlabels = ["Maxima " + lbl, "Minima " + lbl, "Max+Min " + lbl, "Abs-Max " + lbl]
    table = np.vstack(
        (ec["Maximum"][j], ec["Minimum"][j], ec["Max+Min"][j], ec["Abs-Max"][j])
    )
    return rowlabels, table


def _wttab_eventcount(f, res, ec, count_filter, desclen, descfrm, caselen):
    # extrema count
    n = len(res.cases)
    f.write(f"Extrema Count\nFilter: {count_filter}\n\n")
    widths = [desclen, *([caselen] * n)]
    headers = ["Description", *res.cases]
    for j, frm, lbl in zip(
        (0, 1), (f"{{:{caselen}d}}", f"{{:{caselen}.1f}}"), ("Count", "Percent")
    ):
        formats = [descfrm, *([frm] * n)]
        hu_, frm_ = writer.formheader(
            headers, widths, formats, sep=[7, 1], just="c", ulchar="="
        )
        f.write(hu_)
        rowlabels, table = _rowlbls_table(lbl, ec, j)
        writer.vecwrite(f, frm_, rowlabels, table)
        if j == 0:
            f.write("\n")


def _wtxlsx_eventcount(workbook, header, bold, hform, res, ec, name, count_filter):
    # extrema count
    sheet = f"{name} Count"
    worksheet = workbook.add_worksheet(sheet)

    title = header.split("\n")
    for i, ln in enumerate(title):
        worksheet.write(i, 0, ln, bold)

    n = len(title)
    worksheet.write(n, 0, "Extrema Count", bold)
    worksheet.write(n + 1, 0, f"Filter: {count_filter}", bold)
    worksheet.write(n + 1, 1, count_filter)
    n += 2
    ncases = len(res.cases)
    headers = ["Description", *res.cases]
    chart_positions = ((25, 0), (25, 8), (55, 0), (55, 8))
    data_labels = {"category": True, "percentage": True, "separator": "\n"}
    chart_opts = {"x_offset": 25, "y_offset": 10}
    chart_size = {"width": 600, "height": 500}

    for j, frm, lbl in zip((0, 1), ("#,##", "#,##0.0"), ("Count", "Percent")):
        number = workbook.add_format({"num_format": frm})
        worksheet.write_row(n, 0, headers, hform)
        rowlabels, table = _rowlbls_table(lbl, ec, j)

        # write table and pie charts:
        n += 1
        for i, (rowlbl, chpos) in enumerate(zip(rowlabels, chart_positions)):
            worksheet.write(n + i, 0, rowlbl)
            worksheet.write_row(n + i, 1, table[i], number)
            if j == 1:
                chart = workbook.add_chart({"type": "pie"})
                chart.add_series(
                    {
                        "name": rowlbl,
                        "categories": [sheet, n - 1, 1, n - 1, ncases],
                        "values": [sheet, n + i, 1, n + i, ncases],
                        "data_labels": data_labels,
                    }
                )
                chart.set_title({"name": rowlbl})
                chart.set_size(chart_size)
                worksheet.insert_chart(*chpos, chart, chart_opts)
        n += len(rowlabels) + 1

    # adjust column widths
    worksheet.set_column(0, 0, 20)
    worksheet.set_column(1, len(headers) - 1, 14)


@ytools.write_text_file
def _wttab(
    f, header, hu, frm, res, loop_vars, rows, count_filter, desclen, descfrm, caselen
):
    f.write(header)
    ec = {}  # event counter
    for lbl, vals, case_pv, ext, extcases in loop_vars:
        f.write(f"{lbl} Responses\n\n")
        hu_ = hu.replace("Maximum", lbl)
        f.write(hu_)
        ec[lbl] = _event_count(vals, case_pv, count_filter)
        extcases = _add_zero_case(vals, extcases)
        writer.vecwrite(f, frm, rows, res.drminfo.labels, vals, ext, extcases)
        f.write("\n\n")
    _add_max_plus_min(ec)
    _wttab_eventcount(f, res, ec, count_filter, desclen, descfrm, caselen)


def _wtxlsx(workbook, header, headers, res, loop_vars, name, rows, count_filter):
    bold = workbook.add_format({"bold": True})
    hform = workbook.add_format({"bold": True, "align": "center"})
    frm = _get_numform(res.ext, excel=True)
    number = workbook.add_format({"num_format": frm})
    ec = {}
    for lbl, vals, case_pv, ext, extcases in loop_vars:
        worksheet = workbook.add_worksheet(f"{name} {lbl}")
        title = header.split("\n")
        for i, ln in enumerate(title):
            worksheet.write(i, 0, ln, bold)
        h = [i.replace("Maximum", lbl) for i in headers]
        worksheet.write_row(len(title), 0, h, hform)

        ec[lbl] = _event_count(vals, case_pv, count_filter)
        extcases = _add_zero_case(vals, extcases)

        # write table:
        n = len(title) + 1
        labels = res.drminfo.labels
        ncases = vals.shape[1]
        for i, rw in enumerate(rows):
            worksheet.write(n + i, 0, rw)
            worksheet.write(n + i, 1, labels[i])
            worksheet.write_row(n + i, 2, vals[i], number)
            worksheet.write(n + i, 2 + ncases, ext[i], number)
            worksheet.write(n + i, 3 + ncases, extcases[i])

        # adjust column widths and freeze row and col panes
        worksheet.set_column(1, 1, 20)  # description
        worksheet.set_column(2, 3 + ncases, 14)
        worksheet.freeze_panes(n, 2)

    _add_max_plus_min(ec)
    _wtxlsx_eventcount(workbook, header, bold, hform, res, ec, name, count_filter)


def rpttab1(res, filename, title, count_filter=1e-6, name=None):
    """
    Write results tables with bin count information.

    Parameters
    ----------
    res : SimpleNamespace
        Results data structure with attributes `.ext`, `.cases`,
        `.drminfo`, etc (see example in :class:`DR_Results`)
    filename : string or file_like or 1 or None
        If a string that ends with '.xlsx', a Microsoft Excel file is
        written.

        Otherwise, `filename` is either a name of a file, or is a
        file_like object as returned by :func:`open` or
        :class:`io.StringIO`. Input as integer 1 to write to
        stdout. Can also be the name of a directory or None; in these
        cases, a GUI is opened for file selection.
    title : string
        Title for report
    count_filter : scalar; optional
        Filter to use for the bin count; only numbers larger (in
        the absolute value sense) than the filter are counted
    name : string or None; optional
        For '.xlsx' files, this string is used for sheet naming. If
        None and writing an '.xlsx' file, a ValueError is raised.

    Notes
    -----
    The output files contain the maximums, minimums, abs-max
    tables. The extrema value is also included along with the case
    that produced it.

    After those three tables, a table of bin counts is printed
    showing the number of extrema values produced by each event.
    """

    if len(res.drminfo.labels) != res.mx.shape[0]:
        raise ValueError(
            "length of `labels` does not match number of"
            f" rows in `res.mx` (`desc` = {res.drminfo.desc})"
        )

    rows = np.arange(res.mx.shape[0]) + 1
    headers = ["Row", "Description", *res.cases, "Maximum", "Case"]
    header = title + "\n\n" + _get_rpt_headers(res) + "\n"

    amx, amxcase, aext = _get_absmax(res)
    # order of tables: max, min, abs-max with sign:
    loop_vars = (
        ("Maximum", res.mx, np.nanargmax(res.mx, axis=1), res.ext[:, 0], res.maxcase),
        ("Minimum", res.mn, np.nanargmin(res.mn, axis=1), res.ext[:, 1], res.mincase),
        ("Abs-Max", amx, np.nanargmax(abs(amx), axis=1), aext, amxcase),
    )

    if isinstance(filename, xlsxwriter.workbook.Workbook):
        excel = "old"
    elif isinstance(filename, str) and filename.endswith(".xlsx"):
        excel = "new"
    else:
        excel = ""

    if not excel:
        desclen = max(15, len(max(res.drminfo.labels, key=len)))
        caselen = max(13, len(max(res.cases, key=len)))
        n = len(res.cases)
        widths = [6, desclen, *([caselen] * n), caselen, caselen]
        descfrm = f"{{:{desclen:d}}}"
        numform = f"{{:{caselen}.6e}}"
        formats = ["{:6d}", descfrm, *([numform] * n), numform, "{}"]
        hu, frm = writer.formheader(
            headers, widths, formats, sep=[0, 1], just="c", ulchar="="
        )
        _wttab(
            filename,
            header,
            hu,
            frm,
            res,
            loop_vars,
            rows,
            count_filter,
            desclen,
            descfrm,
            caselen,
        )
    else:
        if not name:
            raise ValueError('`name` must be input when writing ".xlsx" files')
        if excel == "new":
            opts = {"nan_inf_to_errors": True}
            with xlsxwriter.Workbook(filename, opts) as workbook:
                _wtxlsx(
                    workbook, header, headers, res, loop_vars, name, rows, count_filter
                )
        else:
            _wtxlsx(filename, header, headers, res, loop_vars, name, rows, count_filter)



# -*- coding: utf-8 -*-
from types import SimpleNamespace
import numpy as np
from pyyeti import expmint


class SolveExp1(object):
    """
    1st order ODE time domain solver based on the matrix exponential.

    This class is for solving: ``yd - A y = f``

    This solver is exact assuming either piece-wise linear or
    piece-wise constant forces. See :class:`SolveExp2` for more
    details on how this algorithm solves the ODE.

    Examples
    --------
    Calculate impulse response of state-space system::

        xd = A @ x + B @ u
        y  = C @ x + D @ u

    where:
        - force = 0's
        - velocity(0) = B

    .. plot::
        :context: close-figs

        >>> from pyyeti import ode
        >>> from pyyeti.ssmodel import SSModel
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> f = 5           # 5 hz oscillator
        >>> w = 2*np.pi*f
        >>> w2 = w*w
        >>> zeta = .05
        >>> h = .01
        >>> nt = 500
        >>> A = np.array([[0, 1], [-w2, -2*w*zeta]])
        >>> B = np.array([[0], [3]])
        >>> C = np.array([[8, -5]])
        >>> D = np.array([[0]])
        >>> F = np.zeros((1, nt), float)
        >>> ts = ode.SolveExp1(A, h)
        >>> sol = ts.tsolve(B @ F, B[:, 0])
        >>> y = C @ sol.d
        >>> fig = plt.figure('Example')
        >>> fig.clf()
        >>> ax = plt.plot(sol.t, y.T,
        ...               label='SolveExp1')
        >>> ssmodel = SSModel(A, B, C, D)
        >>> z = ssmodel.c2d(h=h, method='zoh')
        >>> x = np.zeros((A.shape[1], nt+1), float)
        >>> y2 = np.zeros((C.shape[0], nt), float)
        >>> x[:, 0:1] = B
        >>> for k in range(nt):
        ...     x[:, k+1] = z.A @ x[:, k] + z.B @ F[:, k]
        ...     y2[:, k]  = z.C @ x[:, k] + z.D @ F[:, k]
        >>> ax = plt.plot(sol.t, y2.T, label='discrete')
        >>> leg = plt.legend(loc='best')
        >>> np.allclose(y, y2)
        True

        Compare against scipy:

        >>> from scipy import signal
        >>> ss = ssmodel.getlti()
        >>> tout, yout = ss.impulse(T=sol.t)
        >>> ax = plt.plot(tout, yout, label='scipy')
        >>> leg = plt.legend(loc='best')
        >>> np.allclose(yout, y.ravel())
        True
    """

    def __init__(self, A, h, order=1):
        """
        Instantiates a :class:`SolveExp1` solver.

        Parameters
        ----------
        A : 2d ndarray
            The state-space matrix: ``yd - A y = f``
        h : scalar or None
            Time step or None; if None, the `E`, `P`, `Q` members will
            not be computed.
        order : scalar, optional
            Specify which solver to use:

            - 0 for the zero order hold (force stays constant across
              time step)
            - 1 for the 1st order hold (force can vary linearly across
              time step)

        Notes
        -----
        The class instance is populated with the following members:

        =======   =================================================
        Member    Description
        =======   =================================================
        A         the input `A`
        h         time step
        n         number of total DOF (``A.shape[0]``)
        order     order of solver (0 or 1; see above)
        E, P, Q   output of :func:`pyyeti.expmint.getEPQ`; they are
                  matrices used to solve the ODE
        pc        True if E, P, and Q member have been calculated;
                  False otherwise
        =======   =================================================

        The E, P, and Q entries are used to solve the ODE::

            for j in range(1, nt):
                d[:, j] = E*d[:, j-1] + P*F[:, j-1] + Q*F[:, j]
        """
        if h:
            E, P, Q = expmint.getEPQ(A, h, order)
            self.E = E
            self.P = P
            self.Q = Q
        self.A = A
        self.h = h
        self.order = order
        self.n = A.shape[0]

    def tsolve(self, force, d0=None):
        """
        Solve time-domain 1st order ODE equations.

        Parameters
        ----------
        force : 2d ndarray
            The force matrix; ndof x time
        d0 : 1d ndarray; optional
            Displacement initial conditions; if None, zero ic's are
            used.

        Returns
        -------
        A record (SimpleNamespace class) with the members:

        d : 2d ndarray
            Displacement; ndof x time
        v : 2d ndarray
            Velocity; ndof x time
        h : scalar
            Time-step
        t : 1d ndarray
            Time vector: np.arange(d.shape[1])*h
        """
        force = np.atleast_2d(force)
        if force.shape[0] != self.n:
            raise ValueError(
                f"Force matrix has {force.shape[0]} rows; {self.n} rows are expected"
            )
        nt = force.shape[1]
        d = np.zeros((self.n, nt))  # , float, order='F')
        if d0 is not None:
            d[:, 0] = d0
        else:
            d0 = np.zeros(self.n, float)
        if nt > 1:
            if not self.h:
                raise RuntimeError("instantiate the class with a valid time step.")
            # calc force term outside loop:
            if self.order == 1:
                PQF = self.P @ force[:, :-1] + self.Q @ force[:, 1:]
            else:
                PQF = self.P @ force[:, :-1]
            E = self.E
            for j in range(1, nt):
                d0 = d[:, j] = E @ d0 + PQF[:, j - 1]
            t = self.h * np.arange(nt)
        else:
            t = np.array([0.0])
        return SimpleNamespace(d=d, v=force + self.A @ d, h=self.h, t=t)



# -*- coding: utf-8 -*-
"""
Collection of tools for writing formatted text to files.
"""

import numpy as np
from pyyeti import ytools


def getith(i, args, fncs):
    """
    Return list with i'th value from each input, typically called by
    :func:`vecwrite`.

    Parameters
    ----------
    i : integer
        Specifies which value to extract from each input; starts at 0.
    args : list of variables
        Variable to extract the i'th value from. Must be compatibly
        sized (scalars or vectors of equal length). Strings are
        considered scalars.
    fncs : list of functions
        Same length as args; the function is used to extract the
        i'th item. Call signature:  ith_element_of_a = func(a, i).
        The function must return an iterable of items (eg, list).

    Returns
    -------
    lst : list
        List of the i'th items extracted from each variable in `args`.

    Examples
    --------
    >>> from pyyeti import writer
    >>> import numpy as np
    >>> r = np.array([1.2, 45.])
    >>> s = 'test string'
    >>> i = 5
    >>> v = ['One', 'Two']
    >>> def f(a, i): return [a]
    >>> def f2(a, i): return [a[i]]
    >>> args = [r, s, i, v]
    >>> fncs = [f2, f, f, f2]
    >>> writer.getith(0, args, fncs)
    [1.2, 'test string', 5, 'One']
    >>> writer.getith(1, args, fncs)
    [45.0, 'test string', 5, 'Two']
    """
    lst = []
    for arg, fnc in zip(args, fncs):
        lst.extend(fnc(arg, i))
    return lst


@ytools.write_text_file
def _vecwrite(fout, string, length, args, fncs, postfunc, pfargs, so):
    """Utility routine for :func:`vecwrite`."""
    v = range(length)
    if so is not None:
        v = v[so]
    if postfunc:
        if pfargs is None:
            pfargs = []
        for i in v:
            curargs = getith(i, args, fncs)
            s = postfunc(string.format(*curargs), *pfargs)
            fout.write(s)
    else:
        for i in v:
            curargs = getith(i, args, fncs)
            fout.write(string.format(*curargs))


def vecwrite(f, string, *args, postfunc=None, pfargs=None, so=None):
    """
    Vectorized write.

    Parameters
    ----------
    f : string or file_like or 1 or None
        Either a name of a file, or is a file_like object as returned
        by :func:`open` or :class:`io.StringIO`. Input as integer 1 to
        write to stdout. Can also be the name of a directory or None;
        in these cases, a GUI is opened for file selection.
    string : string
        The formatting string for the write, Python 3 format as in:
        `string`.format(a,b)
    *args : list of variables
        Variables to write. Must be compatibly sized (scalars or
        vectors or numpy arrays of compatible sizes). numpy arrays of
        length 1 are considered scalars. For 2-d numpy arrays, each
        row is written on one line and each element of the row must
        have a conversion specifier. 1-d numpy arrays are treated
        like a column 2-d numpy array. Strings are considered
        scalars.
    postfunc : function or None
        If a function, it is called with the final string (for each
        line) as the argument and it must return a string. The return
        string is what gets output. This can be handy for final string
        substitutions, for example. This input must be named and must
        be after the arguments to be printed; see example.
    pfargs : iterable or None
        If an iterable, contains extra arguments to pass to `postfunc`
        after the string argument. Must be named and after the
        arguments to be printed.
    so : slice object or None
        Allows selection of limited range and custom increment; eg:
        ``slice(0, 10, 2)``. Scalars are not sliced. Must be named and
        after the arguments to be printed.

    Returns
    -------
    None.

    Notes
    -----
    The expected vector length is determined from the first non-scalar
    input. Note that scalar values are repeated automatically as
    necessary.

    Raises
    ------
    ValueError
        When the lengths of print arguments do not match (for
        lengths > 1). Note that the slice object `so` can make
        otherwise incompatible arguments compatible; for example,
        arguments of length 10 and length 100 would be compatible if
        ``so = slice(10)`` (or similar).

    Examples
    --------
    >>> from pyyeti import writer
    >>> import sys
    >>> import numpy as np
    >>> r = np.array([1.2, 45.8])
    >>> s = 'test string'
    >>> i = 5
    >>> v = ['short string', 'a bit longer string']
    >>> frm = '{:3}, {:5.1f}, {:<25}, {}' + chr(10)
    >>> writer.vecwrite(sys.stdout, frm, i, r, v, s)
      5,   1.2, short string             , test string
      5,  45.8, a bit longer string      , test string

    >>> r = np.array([[1.1, 1.2, 1.3], [10.1, 10.2, 10.3]])
    >>> frm = '{:2}, {:=^25} : ' + '  {:6.2f}'*3 + chr(10)
    >>> writer.vecwrite(sys.stdout, frm, i, v, r)
     5, ======short string======= :     1.10    1.20    1.30
     5, ===a bit longer string=== :    10.10   10.20   10.30

    >>> def pf(s):
    ...     return s.replace('0 ', '  ')
    >>> writer.vecwrite(sys.stdout, frm, i, v, r, postfunc=pf)
     5, ======short string======= :     1.1     1.2     1.30
     5, ===a bit longer string=== :    10.1    10.2    10.30

    >>> def pf(s, s_old, s_new):
    ...     return s.replace(s_old, s_new)
    >>> writer.vecwrite(1, frm, i, v, r, postfunc=pf,
    ...                 pfargs=['0 ', '  '])
     5, ======short string======= :     1.1     1.2     1.30
     5, ===a bit longer string=== :    10.1    10.2    10.30
    """

    def _get_scalar(a, i):
        return [a]

    def _get_scalar1(a, i):
        return [a[0]]

    def _get_itemi(a, i):
        return [a[i]]

    def _get_matrow(a, i):
        return a[i]

    length = 1
    fncs = []
    for i, arg in enumerate(args):
        if not isinstance(arg, str) and hasattr(arg, "__len__"):
            if np.ndim(arg) == 2:
                fncs.append(_get_matrow)
                curlen = np.size(arg, 0)
            elif len(arg) == 1:
                fncs.append(_get_scalar1)
                curlen = 1
            else:
                fncs.append(_get_itemi)
                curlen = len(arg)
            if curlen > 1:
                if length > 1:
                    if so is not None:
                        if range(curlen)[so] != range(length)[so]:
                            msg = (
                                "length mismatch with slice object:"
                                f" arg # {i + 1} is incompatible with "
                                "previous args"
                            )
                            raise ValueError(msg)
                    elif curlen != length:
                        msg = (
                            f"length mismatch: arg # {i + 1} has "
                            f"length {curlen}; expected {length} or 1."
                        )
                        raise ValueError(msg)
                length = curlen
        else:
            fncs.append(_get_scalar)
    _vecwrite(f, string, length, args, fncs, postfunc, pfargs, so)


def formheader(headers, widths, formats, sep=(0, 2), just=-1, ulchar="-"):
    """
    Form a nice table header for formatted output via f.write().

    Parameters
    ----------
    headers : list or tuple
        List or tuple of column header strings, eg:
        ['Desc', 'Maximum', 'Time']. Can also be a list of lists (or
        tuples) to support multiple header lines, eg:
        [['Maximum', 'Minimum', 'Time'], ['(lbs)', '(lbs)', '(sec)']]
    widths : iterable
        Iterable of field widths, eg: (25, 10, 8) or [25, 10, 8]. If
        an element in `widths` is < length of corresponding word in a
        header-line, the length of the word is used for that field.
        Note that if this doesn't match with `formats`, the columns
        will not line up nicely.
    formats : list or tuple
        List or tuple of format specifiers for the values in the table,
        eg: ['{:25s}', '{:10f}', '{:8.3f}']
    sep : string, list, tuple, or integer
        Defines 'spacer' in front of each word:

            - if a string, that string is used in front of all headers
            - use a list or tuple of strings for complete control
            - if an integer, that many spaces are used in front of all
              headers
            - use a vector of integers to specify a variable number of
              spaces
            - if len(sep) < len(headers), the last element is used for
              all remaining elements

    just : string or integer or list
        Justification flag or flags for each header string:

            - 'l', 'c', 'r' (or -1, 0, 1) to left, center, or right
              justify headers in their fields
            - can be a list or tuple of len(headers) for complete
              control

    ulchar : string
        Character to use for underlining of headers.

    Returns
    -------
    hu : string
        Contains formatted header string(s) and the underline string.
    f : string
        Final formatting string.

    Examples
    --------
    >>> import numpy as np
    >>> import sys
    >>> from pyyeti import writer
    >>> descs = ['Item 1', 'A different item']
    >>> mx = np.array([[1.2, 2.3], [3.4, 4.5]]) * 1000
    >>> time = np.array([[1.234], [2.345]])
    >>> headers = [['The']*3, ['Descriptions', 'Maximum', 'Time']]
    >>> formats = ['{:<25s}', '{:10.2f}', '{:8.3f}']
    >>> widths  = [25, 10, 8]
    >>> hu, f = writer.formheader(headers, widths, formats,
    ...                           sep=[4, 5, 2], just=0)
    >>> fout = sys.stdout
    >>> if 1:   # just so all output is together
    ...     b = fout.write(hu)
    ...     writer.vecwrite(fout, f, descs, mx, time)
                   The                   The        The
               Descriptions            Maximum      Time
        -------------------------     ----------  --------
        Item 1                           1200.00  2300.000
        A different item                 3400.00  4500.000
    """
    if not isinstance(headers, (list, tuple)):
        raise ValueError("input 'headers' must be a list or tuple")
    if isinstance(headers[0], (list, tuple)):
        length = len(headers[0])
        nheaders = len(headers)
        mxlengths = np.array([len(s) for s in headers[0]])
        for j in range(1, nheaders):
            if len(headers[j]) != length:
                raise ValueError(
                    f"headers[{len(headers[j])}] != length of previous headers"
                )
            for k in range(length):
                mxlengths[k] = max(mxlengths[k], len(headers[j][k]))
    else:
        nheaders = 0
        mxlengths = np.array([len(s) for s in headers])
        length = len(headers)

    if not length == len(formats) == len(widths):
        s = ""
        if isinstance(headers[0], (list, tuple)):
            s = "[*]"
        raise ValueError(
            f"this check failed: ``len(headers{s}) == len(formats) == len(widths)``"
        )

    def strexp(string, width, just):
        if just == -1 or just == "l":
            return string.ljust(width)
        if just == 0 or just == "c":
            return string.center(width)
        return string.rjust(width)

    if isinstance(just, (str, int)):
        just = [just]
    if isinstance(sep, int):
        sep = " " * sep
    if isinstance(sep, str):
        sep = [sep]

    if nheaders > 0:
        h = [""] * nheaders
    else:
        h = ""
    u, f = "", ""
    for j in range(length):
        if j >= len(just):
            cj = just[-1]
        else:
            cj = just[j]
        if j >= len(sep):
            csep = sep[-1]
        else:
            csep = sep[j]
        if isinstance(csep, int):
            csep = " " * csep
        w = max(widths[j], mxlengths[j])
        if nheaders > 0:
            for k in range(nheaders):
                h[k] += csep + strexp(headers[k][j], w, just=cj)
        else:
            h += csep + strexp(headers[j], w, just=cj)
        u += csep + ulchar * w
        f += csep + formats[j]

    if nheaders > 0:
        h = [hj.rstrip() + "\n" for hj in h]
    else:
        h = h.rstrip() + "\n"
    u = u.rstrip() + "\n"
    f = f.rstrip() + "\n"
    return "".join(h) + u, f



# -*- coding: utf-8 -*-
"""Plain Python version of the rainflow algorithm. Numba
(``numba.jit(nopython=True)``) is used if available to achieve speeds
comparable to the compiled-c version.
"""
import numpy as np


def rainflow(peaks, getoffsets=False):
    """
    Rainflow cycle counting in plain Python (slow).

    Parameters
    ----------
    peaks : 1d array-like
        Vector of alternating peaks (as returned by
        :func:`pyyeti.cyclecount.findap`, for example)
    getoffsets : bool; optional
        If True, the tuple ``(rf, os)`` is returned; otherwise, only
        `rf` is returned.

    Returns
    -------
    rf : 2d ndarray
        n x 3 matrix with the rainflow cycle count information
        ``[amp, mean, count]``:

            - amp is the cycle amplitude (half the peak-to-peak range)
            - mean is mean of the cycle
            - count is either 0.5 or 1.0 depending on whether it's
              half or full cycle

    os : 2d ndarray; optional
        n x 2 matrix of cycle offsets ``[start, stop]``. Only returned
        if `getoffsets` is True. The start and stop values are:

            - start is the offset into `peaks` for start of cycle
            - stop is the offset into `peaks` for end of cycle

    Notes
    -----
    This algorithm is derived from reference [#rain2]_. The compiled C
    version is preferred over this one and is very fast (the logic is
    the same).

    References
    ----------
    .. [#rain2] "Standard Practices for Cycle Counting in Fatigue
           Analysis", ASTM E 1049 - 85 (Reapproved 2005).

    Examples
    --------
    Run the example from the ASTM paper:

    >>> from pyyeti.rainflow.py_rain import rainflow
    >>> rainflow([-2, 1, -3, 5, -1, 3, -4, 4, -2])
    array([[ 1.5, -0.5,  0.5],
           [ 2. , -1. ,  0.5],
           [ 2. ,  1. ,  1. ],
           [ 4. ,  1. ,  0.5],
           [ 4.5,  0.5,  0.5],
           [ 4. ,  0. ,  0.5],
           [ 3. ,  1. ,  0.5]])

    With offsets:

    >>> rf, os = rainflow([-2, 1, -3, 5, -1, 3, -4, 4, -2],
    ...                   getoffsets=True)
    >>> rf
    array([[ 1.5, -0.5,  0.5],
           [ 2. , -1. ,  0.5],
           [ 2. ,  1. ,  1. ],
           [ 4. ,  1. ,  0.5],
           [ 4.5,  0.5,  0.5],
           [ 4. ,  0. ,  0.5],
           [ 3. ,  1. ,  0.5]])
    >>> os              # doctest: +ELLIPSIS
    array([[0, 1],
           [1, 2],
           [4, 5],
           [2, 3],
           [3, 6],
           [6, 7],
           [7, 8]]...)
    """
    peaks = np.atleast_1d(peaks)
    L = peaks.size if peaks.ndim == 1 else 0
    if L < 2:
        raise ValueError("`peaks` must be a real vector with length >= 2")
    if getoffsets:
        return _rainflow2(peaks, L)
    return _rainflow1(peaks, L)


def _rainflow1(peaks, L):
    # not getting offsets:
    pts = np.empty(L)
    rf = np.empty((L - 1, 3))
    j = -1
    fullcyclesp1 = 1  # full cycles plus 1
    n = -1
    for k in range(L):
        # /* step 1 from [1]: */
        j += 1
        pts[j] = peaks[k]
        # /* step 2 from [1]: */
        while j > 1:
            # /* step 3 from [1]: */
            Y = abs(pts[j - 2] - pts[j - 1])
            X = abs(pts[j - 1] - pts[j])
            if X < Y:
                break
            if j == 2:
                # /* step 5 from [1]: */
                # /* [count Y as half cycle] */
                n += 1
                rf[n, 0] = Y / 2
                rf[n, 1] = (pts[0] + pts[1]) / 2
                rf[n, 2] = 0.5
                pts[0] = pts[1]  # /* discard j-2 pt */
                pts[1] = pts[2]
                j = 1
            else:
                # /* step 4 from [1]: */
                # /* [count Y as full cycle] */
                fullcyclesp1 += 1
                n += 1
                rf[n, 0] = Y / 2
                rf[n, 1] = (pts[j - 2] + pts[j - 1]) / 2
                rf[n, 2] = 1.0
                pts[j - 2] = pts[j]  # /* discard j-2, j-1 pts */
                j -= 2

    # /* step 6 from [1]: */
    # /* [count all ranges in pts as half cycles] */
    A = pts[0]
    for k in range(j):
        B = pts[k + 1]
        n += 1
        rf[n, 0] = abs(A - B) / 2
        rf[n, 1] = (A + B) / 2
        rf[n, 2] = 0.5
        A = B

    return rf[: L - fullcyclesp1]


def _rainflow2(peaks, L):
    """Utility routine for :func:`rainflow`; returns (rf, os)."""
    pts = np.empty(L)
    rf = np.empty((L - 1, 3))
    j = -1
    fullcyclesp1 = 1  # full cycles plus 1
    n = -1
    cycle_index = np.empty(L, np.int64)
    os = np.empty((L - 1, 2), np.int64)
    for k in range(L):
        # /* step 1 from [1]: */
        j += 1
        pts[j] = peaks[k]
        cycle_index[j] = k
        # /* step 2 from [1]: */
        while j > 1:
            # /* step 3 from [1]: */
            Y = abs(pts[j - 2] - pts[j - 1])
            X = abs(pts[j - 1] - pts[j])
            if X < Y:
                break
            if j == 2:
                # /* step 5 from [1]: */
                # /* [count Y as half cycle] */
                n += 1
                rf[n, 0] = Y / 2
                rf[n, 1] = (pts[0] + pts[1]) / 2
                rf[n, 2] = 0.5
                os[n, 0] = cycle_index[0]
                os[n, 1] = cycle_index[1]
                pts[0] = pts[1]  # /* discard j-2 pt */
                pts[1] = pts[2]
                cycle_index[0] = cycle_index[1]
                cycle_index[1] = cycle_index[2]
                j = 1
            else:
                # /* step 4 from [1]: */
                # /* [count Y as full cycle] */
                fullcyclesp1 += 1
                n += 1
                rf[n, 0] = Y / 2
                rf[n, 1] = (pts[j - 2] + pts[j - 1]) / 2
                rf[n, 2] = 1.0
                os[n, 0] = cycle_index[j - 2]
                os[n, 1] = cycle_index[j - 1]
                pts[j - 2] = pts[j]  # /* discard j-2, j-1 pts */
                cycle_index[j - 2] = cycle_index[j]
                j -= 2

    # /* step 6 from [1]: */
    # /* [count all ranges in pts as half cycles] */
    A = pts[0]
    for k in range(j):
        B = pts[k + 1]
        n += 1
        rf[n, 0] = abs(A - B) / 2
        rf[n, 1] = (A + B) / 2
        rf[n, 2] = 0.5
        os[n, 0] = cycle_index[k]
        os[n, 1] = cycle_index[k + 1]
        A = B

    return rf[: L - fullcyclesp1], os[: L - fullcyclesp1]


try:
    import numba
except ImportError:
    pass
else:
    _rainflow1 = numba.jit(nopython=True)(_rainflow1)
    _rainflow2 = numba.jit(nopython=True)(_rainflow2)



# -*- coding: utf-8 -*-

from types import SimpleNamespace
import warnings
import scipy.linalg as la
import numpy as np
from pyyeti import locate


# FIXME: We need the str/repr formatting used in Numpy < 1.14.
try:
    np.set_printoptions(legacy="1.13")
except TypeError:
    pass


__all__ = [
    "get_su_coef",
    "get_freq_damping",
    "eigss",
    "delconj",
    "addconj",
    "make_A",
    "solvepsd",
]


def get_su_coef(m, b, k, h, rbmodes=None, rfmodes=None):
    """
    Get uncoupled equations of motion integration coefficients.

    Parameters
    ----------
    m : 1d ndarray or None
        Mass; vector (of diagonal), or None (identity assumed)
    b : 1d ndarray
        Damping; vector (of diagonal)
    k : 1d ndarray
        Stiffness; vector (of diagonal)
    h : scalar or None
        Time step; if None, this return returns None.
    rbmodes : 1d ndarray or None; optional
        Index vector for the rigid-body modes; if None, a mode is
        assumed rigid-body if the `k` value is < 0.005.
    rfmodes : 1d ndarray or None; optional
        Index vector for the residual flexibility modes; if None,
        there are no residual flexibility modes.

    Returns
    -------
    coefs : SimpleNamespace with the members:
        ``F, G, A, B, Fp, Gp, Ap, Bp, pvrb, pvrb_damped``

    Notes
    -----
    All entries in `coefs` are 1d ndarrays. Except for `pvrb`, the
    outputs are the integration coefficients. It can handle
    rigid-body, under-damped, critically-damped, and over-damped
    equations. It can also handle rigid-body with damping. The solver
    is exact with the assumption that the forces vary linearly during
    a time step (1st order hold). `pvrb` is a boolean vector with True
    specifying where rigid-body modes are and `pvrb_damped` is the
    same but for the damped rigid-body modes.

    The coefficients are used as follows::

        for j in range(nt-1):
            d[:, j+1] = (F * d[:, j] + G * v[:, j] +
                           A * P[:, j] + B * P[:, j+1])
            v[:, j+1] = (Fp * d[:, j] + Gp * v[:, j] +
                           Ap * P[:, j] + Bp * P[:, j+1])

    where `d` is the displacement, `v` is the velocity, and `P` is the
    applied force.

    Most of the coefficients can be found in the Nastran Theoretical
    Manual [#nast1]_. For the case where ``k = 0`` but ``b != 0``
    (rigid-body with damping ... which is probably unusual), the
    coefficients were computed by hand and confirmed in Python using
    the "sympy" package.

    References
    ----------
    .. [#nast1] 'The NASTRAN Theoretical Manual', Section 11.5,
           NASA-SP-221(06), Jan 01, 1981.
           https://ntrs.nasa.gov/search.jsp?R=19840010609

    See also
    --------
    :class:`SolveUnc`
    """
    if h is None:
        return None
    n = len(b)
    if m is None:
        wo2 = k
        C = b / 2
    else:
        wo2 = k / m
        C = (b / m) / 2
    w2 = wo2 - C ** 2

    if rbmodes is None:
        pvrb = (wo2 < 0.005).astype(int)
    else:
        pvrb = np.zeros(n, int)
        pvrb[rbmodes] = 1
    pvel = 1 - pvrb

    # make a zeros/ones rfmodes
    rfmodes2 = np.zeros(n, int)
    if rfmodes is not None:
        pvel[rfmodes] = 0
        rfmodes2[rfmodes] = 1
    pvel = pvel.astype(bool)

    # check for damped rigid-body equations
    # - from trial and error, it was found that a decent cutoff value
    #   for velocities (to use damping) is when:
    #          1/(b**2*h) < 10**10
    #     or:  b > 10**-5 / sqrt(h)
    # - displacements have a different cutoff. the denominator
    #   is: 1/(b**3*h) < 10**10
    #     or:  b > (1e-10/h)**(1/3)
    #     adjusted to: b > 10*(1e-10/h)**(1/3) by trial and error
    pvrb_damped = (abs(C) > 1e-5 / np.sqrt(h)) & pvrb

    # setup partition vectors for underdamped, critically damped,
    # and overdamped equations
    badrows = None
    if np.any(pvel):
        pvover = np.zeros(n, bool)
        pvcrit = np.zeros(n, bool)
        pvundr = np.zeros(n, bool)
        rat = w2[pvel] / wo2[pvel]
        pvundr[pvel] = rat >= 1.0e-8
        pvcrit[pvel] = abs(rat) < 1.0e-8
        pvover[pvel] = rat <= -1e-8

        if not np.all(rfmodes2 + pvrb + pvundr + pvover + pvcrit == 1):
            badrows = np.nonzero(rfmodes2 + pvrb + pvundr + pvover + pvcrit != 1)[0]
    elif not np.all(rfmodes2 + pvrb == 1):
        badrows = np.nonzero(rfmodes2 + pvrb != 1)[0]

    if badrows is not None:
        msg = f"Partitioning problem. Check settings for mode number(s):\n{badrows}"
        raise ValueError(msg)

    w2 = abs(w2)
    # define the appropriate parameters based on damping
    # ... grab memory and at the same time set the rb equations
    F = pvrb.astype(float)
    G = h * F
    if m is None:
        A = (h * h / 3) * F
        Ap = (h / 2) * F
    else:
        A = (h * h / 3) * F / m
        Ap = (h / 2) * F / m
    B = A / 2
    Fp = np.zeros(n, float)
    Gp = F.copy()
    Bp = Ap.copy()

    if np.any(pvrb_damped):
        pvvelo = pvrb_damped.nonzero()[0]
        pvdisp = (abs(C[pvvelo]) > 10 * (1e-10 / h) ** (1 / 3)).nonzero()[0]
        # F & Fp are correct already
        beta = C[pvvelo] * 2
        ibm = 1 / beta if m is None else 1 / (beta * m[pvvelo])
        ibh = 1 / (beta * h)
        ibbh = 1 / (beta * beta * h)
        ex = np.exp(-beta * h)
        if pvdisp.size:
            # damping coefficients for displacements are only used if
            # damping is bigger ... see note above (otherwise, the rb
            # coefficients are more accurate)
            pv = pvvelo[pvdisp]
            G[pv] = ((1 - ex) / beta)[pvdisp]
            A[pv] = (ibm * ((1 / beta + ibbh) * ex + h / 2 - ibbh))[pvdisp]
            B[pv] = (ibm * (h / 2 - 1 / beta + (1 - ex) * ibbh))[pvdisp]
        Gp[pvvelo] = ex
        Ap[pvvelo] = ibm * (ibh - (1 + ibh) * ex)
        Bp[pvvelo] = ibm * (1 + ibh * (ex - 1))

    if np.any(pvel):
        if np.any(pvundr):
            w = np.sqrt(w2[pvundr])
            cs = np.cos(w * h)
            sn = np.sin(w * h)
            beta = C[pvundr]
            ex = np.exp(-beta * h)
            _wo2 = wo2[pvundr]
            _w2 = w2[pvundr]
            _k = k[pvundr]

            # for displacement:
            F[pvundr] = ex * (cs + (beta / w) * sn)
            G[pvundr] = (ex * sn) / w
            t0 = 1 / (h * _k * w)
            t1 = (_w2 - beta * beta) / _wo2
            t2 = (2 * w * beta) / _wo2
            A[pvundr] = t0 * (ex * ((t1 - h * beta) * sn - (t2 + h * w) * cs) + t2)
            B[pvundr] = t0 * (ex * (-t1 * sn + t2 * cs) + w * h - t2)

            # for velocity:
            Fp[pvundr] = -(_wo2 / w) * ex * sn
            Gp[pvundr] = ex * (cs - (beta / w) * sn)
            Ap[pvundr] = t0 * (ex * ((beta + h * _wo2) * sn + w * cs) - w)
            Bp[pvundr] = t0 * (-ex * (beta * sn + w * cs) + w)

        if np.any(pvcrit):
            beta = C[pvcrit]
            ex = np.exp(-beta * h)
            _wo2 = wo2[pvcrit]
            _k = k[pvcrit]

            # for displacement:
            hbeta = h * beta
            F[pvcrit] = ex * (1 + hbeta)
            G[pvcrit] = h * ex
            t0 = 1 / (h * _k)
            A[pvcrit] = t0 * (
                2 / beta - (1 / beta) * ex * (2 + 2 * hbeta + (hbeta * hbeta))
            )
            B[pvcrit] = (t0 / beta) * (hbeta - 2 + ex * (2 + hbeta))

            # for velocity:
            Fp[pvcrit] = -(beta * beta) * (h * ex)
            Gp[pvcrit] = ex * (1 - hbeta)
            Ap[pvcrit] = t0 * (ex * (1 + hbeta + (hbeta * hbeta)) - 1)
            Bp[pvcrit] = t0 * (1 - ex * (hbeta + 1))

        if np.any(pvover):
            w = np.sqrt(w2[pvover])
            cs = np.cosh(w * h)
            sn = np.sinh(w * h)
            beta = C[pvover]
            ex = np.exp(-beta * h)
            _wo2 = wo2[pvover]
            _w2 = w2[pvover]
            _k = k[pvover]

            # for displacement:
            F[pvover] = ex * (cs + (beta / w) * sn)
            G[pvover] = (ex * sn) / w
            t0 = 1 / (h * _k * w)
            t1 = (_w2 + beta * beta) / _wo2
            t2 = (2 * w * beta) / _wo2
            A[pvover] = t0 * (ex * (-(t1 + h * beta) * sn - (t2 + h * w) * cs) + t2)
            B[pvover] = t0 * (ex * (t1 * sn + t2 * cs) + w * h - t2)

            # for velocity:
            Fp[pvover] = -(_wo2 / w) * ex * sn
            Gp[pvover] = ex * (cs - (beta / w) * sn)
            Ap[pvover] = t0 * (ex * ((beta + h * _wo2) * sn + w * cs) - w)
            Bp[pvover] = t0 * (-ex * (beta * sn + w * cs) + w)

    if rfmodes is not None:
        F[rfmodes] = 0.0
        G[rfmodes] = 0.0
        A[rfmodes] = 0.0
        B[rfmodes] = 1.0 / k[rfmodes]  # from k q = force
        Fp[rfmodes] = 0.0
        Gp[rfmodes] = 0.0
        Ap[rfmodes] = 0.0
        Bp[rfmodes] = 0.0

    return SimpleNamespace(
        F=F,
        G=G,
        A=A,
        B=B,
        Fp=Fp,
        Gp=Gp,
        Ap=Ap,
        Bp=Bp,
        pvrb=pvrb,
        pvrb_damped=pvrb_damped,
    )


def _eigc_dups(lam, tol=1.0e-10):
    """
    Find duplicate complex eigenvalues from state-space formulation.

    Parameters
    ----------
    lam : 1d ndarray
        Vector of complex eigenvalues. Any complex-conjugate pairs
        must be adjacent (this is the normal case for
        :func:`scipy.linalg.eig`).
    tol : scalar; optional
        Tolerance for checking for repeated roots; lam[j] and lam[k]
        are repeated roots if abs(lam[j]-lam[k]) < tol

    Returns
    -------
    lams : 1d ndarray
        Sorted version of lam; real (overdamped) lambdas are first;
        complex lambdas are last. There is no sorting within the real
        lambdas. Conjugates will remain adjacent.
    i : 1d ndarray
        The sorting vector: lams = lam[i]
    dups : 1d ndarray
        Index partition vector for repeated roots; it will be empty
        (`np.array([])`) if there are no repeated roots. For example,
        if only the second and third roots are duplicates of each
        other, `dups` will be `np.array([1, 2])`.

    Notes
    -----
    Input `lam` is the vector of eigenvalues of `A` defined by::

        yd - A y = 0

    and the solution of ud - lam u = 0 is of the form::

        u = k exp(lam t)

    This routine is normally called by :func:`eigss`.
    """
    i = np.argsort(abs(lam.imag), kind="mergesort")
    lams = lam[i]  # order: real then complex w/ conjugates adjacent
    # find repeated roots:
    dups = np.nonzero(locate.find_duplicates(lams, tol))[0]
    return lams, i, dups


def get_freq_damping(lam, suppress_warning=False):
    r"""
    Get frequency and damping from complex eigenvalues

    Parameters
    ----------
    lam : 1d ndarray; shape = (2n,)
        Vector of potentially complex eigenvalues. Pairs are assumed
        to be adjacent. Furthermore, each pair of values in `lam` is
        assumed to be: :math:`-\omega_n (\zeta \pm \sqrt {\zeta^2 -
        1})`. This is the case when `lam` are the eigenvalues of
        :math:`A`, where :math:`A` is setup as shown in :func:`eigss`
        (which has a :math:`-K` in it as opposed to a :math:`+K`).
    suppress_warning : bool; optional
        If True, do not print a warning if this routine finds that
        there are complex eigenvalues without an adjacent
        conjugate. Suppressing the warning can be useful for cases
        where the state-space matrix is complex. In that case, this
        routine may not be that useful, but there's no use in printing
        a warning. See also the notes below.

    Returns
    -------
    wn : 1d ndarray; shape = (n,)
        Vector of natural frequencies (rad/sec) in the same order as
        `lam`. `wn` is always greater than or equal to zero.
    zeta : 1d ndarray; shape = (n,)
        Vector of critical damping ratios:

        ========  ====================================================
        \|zeta\|   description of eigenvalue pairs
        ========  ====================================================
         < 1.0    underdamped; eigenvalues are complex conjugate pairs
         = 1.0    critically damped; eigenvalues are real duplicates
         > 1.0    overdamped; eigenvalues are real and unequal
        ========  ====================================================

        .. note::
            If `zeta` comes out negative, it could mean that the sign
            on `lam` is the opposite of that expected. Or, you have
            negative damping. If `zeta` should be positive for your
            problem, just change the sign on `lam` (and see
            description of the `lam` input).

    Notes
    -----
    The implicit assumption behind this routine is that the equations
    of motion are real (that the state-space matrix as documented in
    :class:`SolveUnc` is real). If that is not true, this routine may
    or may not be helpful, probably depending on how close to real the
    system is. Note that this routine will issue a warning if it finds
    that there are complex eigenvalues without an adjacent conjugate
    (unless `suppress_warning` is True). The rest of the documentation
    assumes the system is real.

    The complex eigenvalue pairs for all cases (underdamped,
    critically damped, and overdamped) are:

    .. math::
        \begin{aligned}
           \lambda_1 &= -\omega_n (\zeta + \sqrt {\zeta^2 - 1}) \\
           \lambda_2 &= -\omega_n (\zeta - \sqrt {\zeta^2 - 1})
        \end{aligned}

    The eigenvalue pairs will be real if :math:`|\zeta| >= 1.0` and
    complex conjugates otherwise.

    The natural frequency :math:`\omega_n` is computed by taking the
    square-root of the product of the eigenvalue pairs. This works for
    underdamped, critically damped, and overdamped cases:

    .. math::
        \omega_n = \sqrt{\lambda_1 \lambda_2}

    Because:

    .. math::
        \begin{aligned}
           \lambda_1 \lambda_2 &=
                \omega_n (\zeta + \sqrt {\zeta^2 - 1})
                \omega_n (\zeta - \sqrt {\zeta^2 - 1}) \\
             &= \omega_n^2 (\zeta^2 - (\zeta^2 - 1) \\
             &= \omega_n^2
        \end{aligned}

    Once :math:`\omega_n` is known, :math:`\zeta` can be computed by
    dividing the sum of the eigenvalue pairs by :math:`2 \omega_n`:

    .. math::
        \begin{aligned}
           \zeta &= -\frac{\lambda_1 + \lambda_2}{2 \omega_n} \\
                 &= \frac{\omega_n (\zeta + \sqrt {\zeta^2 - 1})
                        + \omega_n (\zeta - \sqrt {\zeta^2 - 1})}
                        {2 \omega_n} \\
                 &= \frac{2 \omega_n \zeta}{2 \omega_n} \\
                 &= \zeta
        \end{aligned}

    .. note::
        This routine will give incorrect results without warning if
        any real eigenvalue is not adjacent to its pair. (A complex
        eigenvalue that is not adjacent to its pair will trigger a
        warning.)

    Raises
    ------
        ValueError
            When length of `lam` is not even (eigenvalues must be in
            pairs)

    Examples
    --------
    Set up a diagonal system of equations with known natural
    frequencies and damping. Then, to demo the routine, put the
    equations into state-space form (as shown in :class:`SolveUnc`),
    compute eigenvalues, call this routine, and check results:

    >>> import numpy as np
    >>> import scipy.linalg as la
    >>> from pyyeti import ode
    >>>
    >>> m = np.array([10.0, 11.0, 12.0, 13.0])  # mass
    >>> k = np.array([6.0e5, 7.0e5, 8.0e5, 9.0e5])  # stiffness
    >>> zeta = np.array([0.2, 0.05, 1.0, 2.0])  # percent damping
    >>>
    >>> wn = np.sqrt(k / m)
    >>>
    >>> b = 2.0 * zeta * np.sqrt(k / m) * m  # damping
    >>>
    >>> # solve eigenvalue problem on state-space equations:
    >>> A = ode.make_A(m, b, k)
    >>> lam, phi = la.eig(A)
    >>> wn_extracted, zeta_extracted = ode.get_freq_damping(lam)
    >>>
    >>> # check results:
    >>> i = np.argsort(wn_extracted)
    >>> np.allclose(wn_extracted[i], wn)
    True
    >>> np.allclose(zeta_extracted[i], zeta)
    True

    Check sign convention. The first input follows the sign convention
    described above, so should result in a positive 60% damping. The
    second is the opposite, so should give -60% damping:

    >>> ode.get_freq_damping([-3+4j, -3-4j])
    (array([ 5.]), array([ 0.6]))
    >>> ode.get_freq_damping([3+4j, 3-4j])
    (array([ 5.]), array([-0.6]))
    """
    # find complex conjugate pairs:
    lam = np.atleast_1d(lam)

    lam1 = lam[::2]
    lam2 = lam[1::2]
    if lam1.shape[0] != lam2.shape[0]:
        raise ValueError("`lam` must be even length")

    mult = lam1 * lam2
    add = -(lam1 + lam2)
    if not suppress_warning:
        if (abs(mult.imag).max() > 1e-14) or (abs(add.imag).max() > 1e-14):
            warnings.warn(
                "Eigenvalues pairs in `lam` appear not to be adjacent. Multiplying "
                "and adding pairs resulted in a non-zero imaginary parts:\n"
                f"  Multiply: abs((lam1 * lam2).imag).max() = {abs(mult.imag).max()}\n"
                f"  Add:      abs((lam1 + lam2).imag).max() = {abs(add.imag).max()}",
                RuntimeWarning,
            )

    wn = np.sqrt(abs(mult.real))
    zeta = add.real / (2 * wn)
    return wn, zeta


def eigss(A, delcc):
    r"""Solve complex eigen problem for state-space formulation.

    Parameters
    ----------
    A : 2d ndarray
        The state-space matrix (which doesn't have to be defined as
        below)
    delcc : bool
        If True, delete one of each complex-conjugate pair and put the
        appropriate factor of 2. in ur output (see below). This is
        handy for real time-domain solutions, but not for frequency
        domain solutions. See note below.

    Returns
    -------
    A SimpleNamespace with the members:

    lam : 1d ndarray; complex
        The vector of complex eigenvalues
    ur : 2d ndarray; complex
        Normalized complex right eigenvectors
    ur_inv : 2d ndarray; complex
        Inverse of right eigenvectors
    dups : 1d ndarray
        Index partition vector for repeated roots; it will be empty
        (`np.array([])`) if there are no repeated roots. For example,
        if only the second and third roots are duplicates of each
        other, `dups` will be `np.array([1, 2])`.
    wn : 1d ndarray; real
        Vector of natural frequencies (rad/sec) in the same order as
        `lam` (see :func:`get_freq_damping`)
    zeta : 1d ndarray; real
        Vector of critical damping ratios (see
        :func:`get_freq_damping`)
    eig_success : bool
        True if routine is successful. False if the eigenvectors form
        a singular matrix or they do not diagonalize `A`; in that
        case, ODE solution (if computed) is most likely wrong.

    Notes
    -----
    The typical 2nd order ODE is:

    .. math::
        M \ddot{q} + B \dot{q} + K q = F

    The 2nd order ODE set of equations are transformed into the 1st
    order ODE (see :func:`make_A`):

    .. math::
        \left\{
            \begin{array}{c} \ddot{q} \\ \dot{q} \end{array}
        \right\} - \left[
            \begin{array}{cc} -M^{-1} B & -M^{-1} K \\ I & 0 \end{array}
        \right] \left\{
            \begin{array}{c} \dot{q} \\ q \end{array}
        \right\} = \left\{
            \begin{array}{c} M^{-1} F \\ 0 \end{array} \right\}

    or:

    .. math::
        \dot{y} - A y = w

    When the `M`, `B` and `K` are assembled into the `A` matrix, they
    must not contain any rigid-body modes since the inverse of `ur`
    may not exist, causing the method to fail. If you seen any warning
    messages about a matrix being singular or near singular, the
    method has likely failed. Duplicate roots can also cause trouble,
    so if there are duplicates, check to see if ``ur_inv @ ur`` and
    ``ur_inv @ A @ ur`` are diagonal matrices (if ``not delcc``, these
    would be identity and the eigenvalues, but if `delcc` is True, the
    factor of 2.0 discussed next has the chance to modify that). If
    method fails, see :class:`SolveExp1` or :class:`SolveExp2`.

    For underdamped modes, the complex eigenvalues and modes come in
    complex conjugate pairs. Each mode of a pair yields the same
    solution for real time-domain problems. This routine takes
    advantage of this fact (if `delcc` is True) by chopping out one of
    the pair and multiplying the other one by 2.0 (in `ur`). Then, if
    all modes are underdamped: ``len(lam) = M.shape[0]`` and if no
    modes are underdamped: ``len(lam) = 2*M.shape[0]``.

    See also
    --------
    :func:`make_A`, :class:`SolveUnc`, :func:`get_freq_damping`.

    """
    warn1 = (
        "in :func:`eigss`, the eigenvectors for the state-"
        "space formulation are poorly conditioned (cond={:.3e}).\n"
    )

    warn2 = (
        "Repeated roots detected and equations do not appear "
        "to be diagonalized. Generally, this is a failure "
        "condition.\n"
        "\tMax off-diag / on-diag of `inv(ur) @ A @ ur` = {} / {} = {}\n"
    )

    note = (
        "Solution will likely be inaccurate. "
        "Possible causes/solutions:\n"
        "\tThe partition vector for the rigid-body modes is "
        "incorrect or not set\n"
        "\tThe equations are not in modal space, and the "
        "rigid-body modes cannot be detected -- use the "
        "`pre_eig` option\n"
        "\tUse :class:`SolveExp2` instead for time domain, or\n"
        "\tUse :class:`FreqDirect` instead for frequency domain\n\n"
        "\tSetting `eig_success` attribute to False\n"
    )

    lam, ur = la.eig(A)
    c = np.linalg.cond(ur)
    if c > 1 / np.finfo(float).eps:
        warnings.warn(warn1.format(c) + note, RuntimeWarning)
        eig_success = False
    else:
        eig_success = True
    ur_inv = la.inv(ur)
    lam, i, dups = _eigc_dups(lam)
    ur = ur[:, i]
    ur_inv = ur_inv[i]
    if dups.size:
        uau = ur_inv @ A @ ur
        d = np.diag(uau)
        max_off = abs(np.diag(d) - uau).max()
        max_on = abs(d).max()
        if max_off > 1e-8 * max_on:
            warnings.warn(
                warn2.format(max_off, max_on, max_off / max_on) + note, RuntimeWarning
            )
            eig_success = False

    wn, zeta = get_freq_damping(lam, np.iscomplexobj(A))
    if delcc:
        lam, ur, ur_inv, dups = delconj(lam, ur, ur_inv, dups)
    return SimpleNamespace(
        lam=lam,
        ur=ur,
        ur_inv=ur_inv,
        dups=dups,
        wn=wn,
        zeta=zeta,
        eig_success=eig_success,
    )


def delconj(lam, ur, ur_inv, dups):
    """
    Delete one eigenvalue/eigenvector from of each pair of complex
    conjugates.

    Parameters
    ----------
    lam : 1d ndarray
        The vector of complex eigenvalues
    ur : 2d ndarray
        Normalized complex right eigenvectors
    ur_inv : 2d ndarray
        Inverse of right eigenvectors
    dups : 1d array_like
        Index partition vector for repeated roots; it will be empty
        (`np.array([])`) if there are no repeated roots. For example,
        if only the second and third roots are duplicates of each
        other, `dups` will be `np.array([1, 2])`.

    Returns
    -------
    lam1 : 1d ndarray; complex
        Trimmed vector of complex eigenvalues
    ur1 : 2d ndarray; complex
        Trimmed normalized complex right eigenvectors; columns may be
        trimmed
    ur_inv1 : 2d ndarray; complex
        Trimmed inverse of right eigenvectors; rows may be trimmed
    dups1 : 1d ndarray
        Version of input `dups` for the trimmed variables.

    Notes
    -----
    This function is typically called via :func:`eigss`. If there are
    any values in `lam` that are not part of a complex-conjugate pair,
    this routine does nothing.

    :func:`delconj` can safely be called even if modes were already
    deleted. In this case, outputs will be the same as the inputs.
    """
    # delete the negatives (imaginary part) of each comp-conj pair:
    neg = lam.imag < 0.0
    if np.any(neg):
        # see if lambda's are all comp-conj pairs; if not, do nothing
        posi = np.nonzero(lam.imag > 0.0)[0]
        negi = np.nonzero(neg)[0]
        if (
            posi.size == negi.size
            and np.all(abs(posi - negi) == 1)
            and np.all(lam[posi] == np.conj(lam[negi]))
        ):
            # so, keep reals and positives of each complex conj. pair:
            keep = np.logical_not(neg)
            lam = lam[keep]
            ur = ur[:, keep]
            ur_inv = ur_inv[keep]
            # update dups:
            dups = np.atleast_1d(dups)
            if dups.size > 0:
                ndof = ur.shape[0]
                temp = np.zeros(ndof, bool)
                temp[dups] = True
                temp = temp[keep]
                dups = np.nonzero(temp)[0]
            # put factor of 2 in ur for recovery (because only one of
            # the complex conjugate pairs were solved for)
            pv = lam.imag > 0.0
            if np.any(pv):
                ur[:, pv] *= 2.0
    return lam, ur, ur_inv, dups


def addconj(lam, ur, ur_inv):
    """
    Add back in the missing complex-conjugate mode

    Parameters
    ----------
    lam : 1d ndarray
        The vector of complex eigenvalues
    ur : 2d ndarray
        Normalized complex right eigenvectors
    ur_inv : 2d ndarray
        Inverse of right eigenvectors

    Returns
    -------
    lam1 : 1d ndarray; complex
        The vector of complex eigenvalues (complete set)
    ur1 : 2d ndarray; complex
        Complete set of normalized complex right eigenvectors; will be
        square
    ur_inv1 : 2d ndarray; complex
        Complete set of inverse of right eigenvectors; will be square

    Notes
    -----
    This routine adds the missing one of each complex-conjugate pair
    from state-space complex eigensolution and divides the other one
    by 2 in ur.

    :func:`addconj` can safely be called even if modes were already
    added back in (or if they were never deleted).

    Though unlikely, :func:`addconj` could be fooled into adding
    inappropriate modes if modes were deleted in a different manner
    than how :func:`eigss` does it. This routine does some checks to
    try to ensure that the inputs have been processed as expected:

    - checks for all positive imaginary parts in lam (:func:`eigss`
      deletes the negative conjugates)
    - check for the factor of 2 (see above) in ur
    - check that `ur1` and `ur_inv1` will be square after adding in
      the modes
    - check that all underdamped modes are sorted last (as
      :func:`eigss` has them)

    See also
    --------
    :func:`eigss`, :class:`SolveUnc`, :class:`SolveExp2`
    """
    conj2 = np.nonzero(lam.imag > 0.0)[0]
    if ur.shape[0] > ur.shape[1] and conj2.size > 0:
        if np.any(lam.imag < 0.0):
            return lam, ur, ur_inv
        two = ur_inv[conj2[0]] @ ur[:, conj2[0]]
        if abs(two - 2.0) > 1e-13:
            raise ValueError(
                "factor of 2.0 seems to be missing: "
                f"error on first underdamped mode = {abs(two - 2.0)}"
            )
        n = len(lam) + len(conj2)
        if n != ur_inv.shape[1] or n != ur.shape[0]:
            raise ValueError(
                "ur and/or ur_inv will not be square after adding missing modes"
            )
        reals = np.nonzero(lam.imag == 0.0)[0]
        if len(reals) > 0 and np.max(reals) > np.min(conj2):
            raise ValueError("not all underdamped are last")
        conj2_new = conj2 + np.arange(1, len(conj2) + 1)
        conj1_new = conj2_new - 1
        lam1 = np.zeros(n, complex)
        ur_inv1 = np.zeros((n, n), complex)
        ur1 = np.zeros((n, n), complex)
        if reals.size > 0:
            lam1[reals] = lam[reals]
            ur1[:, reals] = ur[:, reals]
            ur_inv1[reals] = ur_inv[reals]
        lam1[conj1_new] = lam[conj2]
        lam1[conj2_new] = np.conj(lam[conj2])
        ur1[:, conj1_new] = ur[:, conj2] / 2.0
        ur1[:, conj2_new] = np.conj(ur[:, conj2]) / 2.0
        ur_inv1[conj1_new] = ur_inv[conj2]
        ur_inv1[conj2_new] = np.conj(ur_inv[conj2])
        return lam1, ur1, ur_inv1
    return lam, ur, ur_inv


def make_A(M, B, K):
    r"""
    Setup the state-space matrix from mass, damping and stiffness.

    Parameters
    ----------
    M : 1d or 2d ndarray or None
        Mass; vector (of diagonal), or full; if None, mass is assumed
        identity
    B : 1d or 2d ndarray
        Damping; vector (of diagonal), or full
    K : 1d or 2d ndarray
        Stiffness; vector (of diagonal), or full

    Returns
    -------
    A : 2d ndarray
        The state-space matrix defined as shown below

    Notes
    -----
    The typical 2nd order ODE is:

    .. math::
        M \ddot{q} + B \dot{q} + K q = F

    The 2nd order ODE set of equations are transformed into the
    1st order ODE:

    .. math::
        \left\{
            \begin{array}{c} \ddot{q} \\ \dot{q} \end{array}
        \right\} - \left[
            \begin{array}{cc} -M^{-1} B & -M^{-1} K \\ I & 0 \end{array}
        \right] \left\{
            \begin{array}{c} \dot{q} \\ q \end{array}
        \right\} = \left\{
            \begin{array}{c} M^{-1} F \\ 0 \end{array} \right\}

    or:

    .. math::
        \dot{y} - A y = w

    When the `M`, `B` and `K` are assembled into the `A` matrix, they
    must not contain any rigid-body modes. See :func:`eigss`.

    See also
    --------
    :func:`eigss`, :class:`SolveUnc`, :class:`SolveExp2`
    """
    Atype = float
    if M is None:
        B, K = np.atleast_1d(B, K)
        if np.iscomplexobj(B) or np.iscomplexobj(K):
            Atype = complex
    else:
        M, B, K = np.atleast_1d(M, B, K)
        if np.iscomplexobj(M) or np.iscomplexobj(B) or np.iscomplexobj(K):
            Atype = complex
    n = K.shape[0]
    A = np.zeros((2 * n, 2 * n), Atype)
    v1 = range(n)
    v2 = range(n, 2 * n)
    if B.ndim == 2:
        A[:n, :n] = -B
    else:
        A[v1, v1] = -B
    if K.ndim == 2:
        A[:n, n:] = -K
    else:
        A[v1, v2] = -K
    A[v2, v1] = 1.0
    if M is not None:
        if M.ndim == 1:
            A[:n] = (1.0 / M).reshape(-1, 1) * A[:n]
        else:
            A[:n] = la.solve(M, A[:n])
    return A


def solvepsd(fs, forcepsd, t_frc, freq, drmlist, rbduf=1.0, elduf=1.0, **kwargs):
    """
    Solve equations of motion in frequency domain with uncorrelated
    PSD forces.

    See also :func:`pyyeti.cla.DR_Results.solvepsd` for a very similar
    routine, but one that is designed for use within the pyYeti
    "coupled loads analysis paradigm" (where the classes defined in
    :mod:`pyyeti.cla` are used).

    Parameters
    ----------
    fs : class instance
        An instance of :class:`SolveUnc` or :class:`FreqDirect` (or
        similar ... must have `.fsolve` method)
    forcepsd : 2d array_like
        The matrix of force psds; each row is a force PSD
    t_frc : 2d array_like
        Transform to put `forcepsd` into the coordinates of the
        equations of motion: ``t_frc @ forcepsd``. Commonly, `t_frc`
        is simply the transpose of a row-partition of the mode shape
        matrix (phi) and the conversion of `forcepsd` is from physical
        space to modal space. In that case, the row-partition is from
        the full set down to just the forced DOF. However, `t_frc` can
        also have force mappings (as from the TLOAD capability in
        Nastran); in that case, ``t_frc = phi.T @
        mapping_vectors``. In any case, the number of columns in
        `t_frc` is the number of rows in `forcepsd`: ``t_frc.shape[1]
        == forcepsd.shape[0]``
    freq : 1d array_like
        Frequency vector at which solution will be computed;
        ``len(freq) = cols(forcepsd)``
    drmlist : list_like
        List of lists (or similar) of any number of data recovery
        matrix quadruples (in the order typically used to write
        equations of motion)::

            [
                [drma1, drmv1, drmd1, drmf1],
                [drma2, drmv2, drmd2, drmf2],
                ...
            ]

        To not use a particular drm, set it to None. For example, to perform
        these 3 types of data recovery::

                acce = atm*a
                disp = dtm*d
                loads = ltma*a + ltmv*v + ltmd*d + ltmf*f

        `drmlist` would be::

              [[atm, None, None, None],
               [None, None, dtm, None],
               [ltma, ltmv, ltmd, ltmf]]

    rbduf : scalar; optional
        Rigid-body uncertainty factor
    elduf : scalar; optional
        Dynamic uncertainty factor
    **kwargs : keyword arguments for ``fs.fsolve``; optional
        Currently, there are two arguments available:

        ============  ============================================
          argument    brief description
        ============  ============================================
        incrb         specifies how to handle rigid-body responses
        rf_disp_only  specifies how to handle residual-flexibility
                      modes
        ============  ============================================

        See, for example, :func:`SolveUnc.fsolve`.

    Returns
    -------
    rms : list
        List of vectors (corresponding to `drmlist`) of rms values of
        all data recovery rows; # of rows of each vector = # rows in
        each drm pair
    psd : list
        List of matrices (corresponding to `drmlist`) of PSD responses
        for all data recovery rows::

               # rows in each PSD = # rows in DRM
               # cols in each PSD = len(freq)

    Notes
    -----
    This routine first calls ``fs.fsolve`` to solve the modal
    equations of motion. Then, it scales the responses by the
    corresponding PSD input force. All PSD responses are summed
    together for the overall response. For example::

        resp_psd = 0
        for i in range(forcepsd.shape[0]):
            # solve for unit frequency response function:
            unitforce = np.ones((1, len(freq)))
            genforce = t_frc[:, i:i+1] @ unitforce
            sol = fs.fsolve(genforce, freq, incrb)
            frf = (drma @ sol.a
                   + drmv @ sol.v
                   + drmd @ sol.d
                   + drmf[:, [i]] @ unitforce)
            resp_psd += forcepsd[i] * abs(frf)**2

    In that example, the data recovery uses all four drms. Also, the
    looping over the `drmlist` is not included for simplicity.

    Examples
    --------
    .. plot::
        :context: close-figs

        >>> from pyyeti import ode
        >>> import numpy as np
        >>> m = np.array([10., 30., 30., 30.])        # diag mass
        >>> k = np.array([0., 6.e5, 6.e5, 6.e5])      # diag stiffness
        >>> zeta = np.array([0., .05, 1., 2.])        # % damping
        >>> b = 2.*zeta*np.sqrt(k/m)*m                # diag damping
        >>> freq = np.arange(.1, 35, .1)              # frequency
        >>> forcepsd = 10000*np.ones((4, freq.size))  # PSD forces
        >>> fs = ode.SolveUnc(m, b, k)
        >>> atm = np.eye(4)    # recover modal accels
        >>> t_frc = np.eye(4)  # assume forces already modal
        >>> drms = [[atm, None, None, None]]
        >>> rms, psd = ode.solvepsd(fs, forcepsd, t_frc, freq,
        ...                         drms)

        The rigid-body results should be 100.0 g**2/Hz flat;
        rms = np.sqrt(100*34.8)

        >>> np.allclose(100., psd[0][0])
        True
        >>> np.allclose(np.sqrt(3480.), rms[0][0])
        True

        Plot the four accelerations PSDs:

        >>> import matplotlib.pyplot as plt
        >>> fig = plt.figure('Example', figsize=[8, 8])
        >>> fig.clf()
        >>> labels = ['Rigid-body', 'Underdamped',
        ...           'Critically Damped', 'Overdamped']
        >>> for j, name in zip(range(4), labels):
        ...     _ = plt.subplot(4, 1, j+1)
        ...     _ = plt.plot(freq, psd[0][j])
        ...     _ = plt.title(name)
        ...     _ = plt.ylabel(r'Accel PSD ($g^2$/Hz)')
        ...     _ = plt.xlabel('Frequency (Hz)')
        >>> fig.tight_layout()
    """
    ndrms = len(drmlist)
    forcepsd, t_frc = np.atleast_2d(forcepsd, t_frc)
    freq = np.atleast_1d(freq)
    rpsd, cpsd = forcepsd.shape
    unitforce = np.ones((1, cpsd))
    psd = [0.0] * ndrms
    rms = [0.0] * ndrms

    if t_frc.shape[1] != rpsd:
        raise ValueError(
            "`forcepsd` and `t_frc` are incompatibly "
            f"sized: {forcepsd.shape} vs {t_frc.shape}"
        )

    for i in range(rpsd):
        # solve for unit frequency response function for i'th force:
        genforce = t_frc[:, i : i + 1] @ unitforce
        sol = fs.fsolve(genforce, freq, **kwargs)
        if rbduf != 1.0:
            sol.a[fs.rb] *= rbduf
            sol.v[fs.rb] *= rbduf
            sol.d[fs.rb] *= rbduf
        if elduf != 1.0:
            sol.a[fs.el] *= elduf
            sol.v[fs.el] *= elduf
            sol.d[fs.el] *= elduf
        for j, (drma, drmv, drmd, drmf) in enumerate(drmlist):
            frf = 0.0
            if drma is not None:
                frf += drma @ sol.a
            if drmv is not None:
                frf += drmv @ sol.v
            if drmd is not None:
                frf += drmd @ sol.d
            if drmf is not None:
                frf += drmf[:, i : i + 1] @ unitforce
            psd[j] += forcepsd[i] * abs(frf) ** 2

    # compute area under curve:
    freqstep = np.diff(freq)
    for j in range(ndrms):
        sumpsd = psd[j][:, :-1] + psd[j][:, 1:]
        rms[j] = np.sqrt(np.sum((freqstep * sumpsd), axis=1) / 2)
    return rms, psd



# -*- coding: utf-8 -*-
from types import SimpleNamespace
import warnings
import scipy.linalg as la
import numpy as np
from pyyeti import ytools


class _BaseODE:
    """
    Base class for time and frequency domain equations of motion
    solvers.

    This class is abstract-like, but not a subclass of abc.ABC because
    I don't want to require that all subclasses implement all the
    "abstract" methods. Raising NotImplementedError seems appropriate
    here.
    """

    def tsolve(self):
        """'Abstract' method to solve time-domain equations"""
        raise NotImplementedError

    def fsolve(self):
        """'Abstract' method to solve frequency-domain equations"""
        raise NotImplementedError

    def generator(self):
        """
        'Abstract' method to get Python "generator" version of the ODE
        solver. This is to interactively solve (or re-solve) one step
        at a time.
        """
        raise NotImplementedError

    def get_f2x(self):
        """
        'Abstract' method to get the force to displacement transform
        used in Henkel-Mar simulations.
        """
        raise NotImplementedError

    def finalize(self, get_force=False):
        """
        Finalize time-domain generator solution.

        Parameters
        ----------
        get_force : bool; optional
            If True, the `force` entry will be included in the
            returned data structure.

        Returns
        -------
        A record (SimpleNamespace class) with the members:

        d : 2d ndarray
            Displacement; ndof x time
        v : 2d ndarray
            Velocity; ndof x time
        a : 2d ndarray
            Acceleration; ndof x time
        h : scalar
            Time-step or None
        t : 1d ndarray
            Time vector: np.arange(d.shape[1])*h
        force : 2d ndarray; optional
            Force; ndof x time. Only included if `get_force` is True.
        """
        d, v, a, f = self._d, self._v, self._a, self._force
        del self._d, self._v, self._a, self._force
        self._calc_acce_kdof(d, v, a, f)
        sol = self._solution(d, v, a)
        if get_force:
            sol.force = f
        return sol

    #
    # Utility routines follow:
    #
    def _solution(self, d, v, a):
        """Returns SimpleNamespace object with d, v, a, h, t"""
        if self.h:
            t = self.h * np.arange(a.shape[1])
        else:
            t = np.array([0.0])
        if self.pre_eig:
            d = self.phi @ d
            v = self.phi @ v
            a = self.phi @ a
        return SimpleNamespace(d=d, v=v, a=a, h=self.h, t=t)

    def _solution_freq(self, d, v, a, freq):
        """Returns SimpleNamespace object with d, v, a, freq"""
        if self.pre_eig:
            d = self.phi @ d
            v = self.phi @ v
            a = self.phi @ a
        return SimpleNamespace(d=d, v=v, a=a, f=freq)

    def _mk_slice(self, pv):
        """Convert index partition vector to slice object:
        ``start:stop``. Raises ValueError if `pv` cannot be converted
        to this type of slice object."""
        if pv.size == 0:
            return slice(0, 0)
        # if pv.size == pv[-1]+1 - pv[0]:
        if np.all(np.diff(pv) == 1):
            return slice(pv[0], pv[-1] + 1)
        raise ValueError("invalid partition vector for conversion to slice")

    def _mk_slices(self):  # , dorbel):
        """Convert index partition vectors to slice objects and sets
        ``slices=True`` if successful."""
        # if not dorbel:
        #     print("Setting dorbel to True for testing...")
        #     dorbel = True
        try:
            nonrf = self._mk_slice(self.nonrf)
            rf = self._mk_slice(self.rf)
            kdof = self._mk_slice(self.kdof)
            # if dorbel:
            rb = self._mk_slice(self.rb)
            el = self._mk_slice(self.el)
            _rb = self._mk_slice(self._rb)
            _el = self._mk_slice(self._el)
        except ValueError:
            self.slices = False
        else:
            self.nonrf = nonrf
            self.rf = rf
            self.kdof = kdof
            # if dorbel:
            self.rb = rb
            self.el = el
            self._rb = _rb
            self._el = _el
            self.slices = True

    def _chk_diag_part(self, m, b, k, cd_as_force=False):
        """Checks for all-diagonal and partitions based on rf modes"""
        krf = None
        cdforces = False

        unc = 0
        if m is None or m.ndim == 1 or (m.ndim == 2 and ytools.isdiag(m)):
            unc += 1

        if b.ndim == 1 or (b.ndim == 2 and ytools.isdiag(b)):
            unc += 1
        elif cd_as_force:
            unc += 1
            cdforces = True

        if k.ndim == 1 or (k.ndim == 2 and ytools.isdiag(k)):
            unc += 1

        if unc == 3:
            unc = True
            if m is not None and m.ndim == 2:
                m = np.diag(m).copy()
            if b.ndim == 2:
                bd = np.diag(b).copy()
                if cdforces:
                    bo = b.copy()
                    i = np.arange(bo.shape[0])
                    bo[i, i] = 0.0  # off diagonal damping
                b = bd
            if k.ndim == 2:
                k = np.diag(k).copy()
            if self.rfsize:
                if m is not None:
                    m = m[self.nonrf]
                b = b[self.nonrf]
                if cdforces:
                    bo = bo[np.ix_(self.nonrf, self.nonrf)]
                krf = k[self.rf]
                k = k[self.nonrf]
        else:
            cdforces = False
            unc = False
            if m is not None and m.ndim == 1:
                m = np.diag(m)
            if b.ndim == 1:
                b = np.diag(b)
            if k.ndim == 1:
                k = np.diag(k)
            if self.rfsize:
                pvrf = np.ix_(self.rf, self.rf)
                pvnonrf = np.ix_(self.nonrf, self.nonrf)
                if m is not None:
                    m = m[pvnonrf]
                b = b[pvnonrf]
                krf = k[pvrf]
                k = k[pvnonrf]
        self.m = m
        self.b = b
        self.cdforces = cdforces
        if cdforces:
            self.bo = bo
        self.k = k
        self.krf = krf
        self.unc = unc

    def _inv_krf(self):
        """Decompose the krf matrix"""
        if self.rfsize:
            krf = self.krf
            if self.unc:
                ikrf = (1.0 / krf).reshape(-1, 1)
                c = abs(krf).max() / abs(krf).min()
            else:
                ikrf = la.lu_factor(krf)
                c = np.linalg.cond(krf)
            if c > 1 / np.finfo(float).eps:
                msg = (
                    "the residual-flexibility part of the stiffness is poorly "
                    f"conditioned (cond={c:.3e}). Displacements will likely "
                    "be inaccurate."
                )
                warnings.warn(msg, RuntimeWarning)
            self.ikrf = ikrf

    def _get_inv_m(self, m):
        """Decompose the mass matrix"""
        if self.unc:
            invm = (1.0 / m).reshape(-1, 1)
            c = abs(m).max() / abs(m).min()
        else:
            invm = la.lu_factor(m)
            c = np.linalg.cond(m)
        if c > 1 / np.finfo(float).eps:
            msg = (
                f"the mass matrix is poorly conditioned (cond={c:.3e}). Solution "
                "will likely be inaccurate."
            )
            warnings.warn(msg, RuntimeWarning)
        return invm

    def _inv_m(self):
        """Decompose the mass matrix"""
        if self.m is not None and self.ksize:
            self.invm = self._get_inv_m(self.m)

    def _inv_mrb(self):
        """Decompose the rigid-body part of the mass matrix"""
        if self.m is not None and self.rbsize:
            if self.unc:
                mrb = self.m[self.rb]
            else:
                mrb = self.m[np.ix_(self.rb, self.rb)]
            self.imrb = self._get_inv_m(mrb)

    def _assert_square(self, n, m, b, k):
        if m is not None:
            name = ("mass", "damping", "stiffness")
            mats = (m, b, k)
        else:
            name = ("damping", "stiffness")
            mats = (b, k)
        any_2d = False
        for i, mat in enumerate(mats):
            if mat.ndim == 2:
                any_2d = True
                if mat.shape[0] != mat.shape[1]:
                    raise ValueError(f"{name[i]} matrix is non-square!")
                if mat.shape[0] != n:
                    raise ValueError(
                        f"{name[i]} matrix has a different number of rows than "
                        "the stiffness!"
                    )
            elif mat.ndim == 1:
                if mat.shape[0] != n:
                    raise ValueError(
                        f"length of {name[i]} diagonal is not compatible with "
                        "the stiffness!"
                    )
            else:
                raise ValueError(f"{name[i]} has more than 2 dimensions!")
        return any_2d

    def _do_pre_eig(self, m, b, k):
        """Do a "pre" eigensolution to put system in modal space"""
        err = False
        if k.ndim == 1:
            k = np.diag(k)
        if m is None:
            w, u = la.eigh(k)
            kdiag = u.T @ k @ u
            if not ytools.isdiag(kdiag):
                err = True
        else:
            if m.ndim == 1:
                m = np.diag(m)
            w, u = la.eigh(k, m)
            kdiag = u.T @ k @ u
            mdiag = u.T @ m @ u
            if not ytools.isdiag(kdiag) or not ytools.isdiag(mdiag):
                err = True
        if err:
            raise ValueError(
                "`pre_eig` option failed to "
                "diagonlized the mass and/or "
                "stiffness. Check "
                "for symmetric/hermitian stiffness "
                "and positive-definite mass"
            )
        self.pre_eig = True
        self.phi = u
        if m is not None:
            m = np.diag(mdiag).copy()
        k = np.diag(kdiag).copy()
        if b.ndim == 1:
            b = (u.T * b) @ u
        else:
            b = u.T @ b @ u
        return m, b, k

    def _ensure_index_type(self, pv):
        pv = np.atleast_1d(pv)
        if np.issubdtype(pv.dtype, np.bool_):
            return np.nonzero(pv)[0]
        return pv

    def _common_precalcs(self, m, b, k, h, rb, rf, pre_eig=False, cd_as_force=False):
        systype = float
        self.mid = id(m)
        self.bid = id(b)
        self.kid = id(k)
        if m is None:
            b, k = np.atleast_1d(b, k)
            if np.iscomplexobj(b) or np.iscomplexobj(k):
                systype = complex
        else:
            m, b, k = np.atleast_1d(m, b, k)
            if np.iscomplexobj(m) or np.iscomplexobj(b) or np.iscomplexobj(k):
                systype = complex
        n = k.shape[0]
        any_2d = self._assert_square(n, m, b, k)
        if pre_eig and any_2d:
            m, b, k = self._do_pre_eig(m, b, k)
        else:
            self.pre_eig = False
        nonrf = np.ones(n, bool)
        if rf is None or (isinstance(rf, list) and not rf):
            rf = np.array([], bool)
        else:
            # # rf = np.ix_(np.atleast_1d(rf))[0]
            # rf = np.atleast_1d(rf)
            rf = self._ensure_index_type(rf)
        nonrf[rf] = False
        nonrf = np.nonzero(nonrf)[0]
        self.n = n
        self.h = h
        self.rf = rf
        self.nonrf = nonrf
        self.rfsize = rf.size
        self.nonrfsz = nonrf.size
        self.kdof = nonrf
        self.ksize = nonrf.size
        self._chk_diag_part(m, b, k, cd_as_force)
        self._make_rb_el(rb)
        self._inv_krf()
        self.systype = systype

    def _make_rb_el(self, rb):
        """
        - rb and el are relative to full size
        - _rb and _el are relative to the non RF part
        """
        if rb is None:
            if self.ksize:
                tol = 0.005
                if self.unc:
                    _rb = np.nonzero(abs(self.k) < tol)[0]
                else:
                    _rb = (
                        (abs(self.k).max(axis=0) < tol)
                        & (abs(self.k).max(axis=1) < tol)
                        & (abs(self.b).max(axis=0) < tol)
                        & (abs(self.b).max(axis=1) < tol)
                    ).nonzero()[0]
                rb = np.zeros(self.n, bool)
                rb[self.nonrf[_rb]] = True
                rb = np.nonzero(rb)[0]
            else:
                rb = _rb = np.array([], bool)
        elif isinstance(rb, list) and not rb:
            rb = _rb = np.array([], bool)
        else:
            # # rb = np.ix_(np.atleast_1d(rb))[0]
            # rb = np.atleast_1d(rb)
            rb = self._ensure_index_type(rb)
            vec = np.zeros(self.n, bool)
            vec[rb] = True
            _rb = np.nonzero(vec[self.nonrf])[0]
        _el = np.ones(self.ksize, bool)
        _el[_rb] = False
        _el = np.nonzero(_el)[0]
        el = np.zeros(self.n, bool)
        el[self.nonrf[_el]] = True
        el = np.nonzero(el)[0]
        # _rb, _el are relative to non-rf part
        self.rb = rb
        self._rb = _rb
        self.el = el
        self._el = _el
        self.rbsize = rb.size
        self.elsize = el.size

    def _build_A(self):
        """Builds the A matrix: yd - A y = f"""
        n = self.k.shape[0]
        A = np.zeros((2 * n, 2 * n), self.systype)
        v1 = range(n)
        v2 = range(n, 2 * n)
        if self.unc:
            A[v1, v1] = -self.b
            A[v1, v2] = -self.k
        else:
            A[:n, :n] = -self.b
            A[:n, n:] = -self.k
        A[v2, v1] = 1.0
        if self.m is not None:
            if self.unc:
                A[:n] *= self.invm
            else:
                A[:n] = la.lu_solve(self.invm, A[:n], check_finite=False)
        return A

    def _alloc_dva(self, nt, istime):
        ORDER = "F"
        n = self.ksize
        if istime:
            if nt > 1 and n > 0 and not self.pc:
                raise RuntimeError(
                    f"rerun `{type(self).__name__}` with a valid time step."
                )
            d = np.zeros((self.n, nt), self.systype, order=ORDER)
            v = np.zeros((self.n, nt), self.systype, order=ORDER)
            a = np.zeros((self.n, nt), self.systype, order=ORDER)
        else:
            d = np.zeros((self.n, nt), complex, order=ORDER)
            v = np.zeros((self.n, nt), complex, order=ORDER)
            a = np.zeros((self.n, nt), complex, order=ORDER)
        return d, v, a

    def _init_dv(self, d, v, d0, v0, F0, static_ic):
        if d0 is not None:
            d[self.nonrf, 0] = d0[self.nonrf]
        elif static_ic and self.elsize and F0[self.el].any():
            # el is relative to full size
            # _el is relative to "k" size
            if self.unc:
                d0 = la.lstsq(np.diag(self.k[self._el]), F0[self.el])
                d[self.el, 0] = d0[0]
            else:
                if self.slices:
                    d0 = la.lstsq(self.k[self._el, self._el], F0[self.el])
                else:
                    d0 = la.lstsq(self.k[np.ix_(self._el, self._el)], F0[self.el])
                d[self.rb, 0] = 0.0
                d[self.el, 0] = d0[0]
        if v0 is not None:
            v[self.nonrf, 0] = v0[self.nonrf]

    def _set_initial_cond(self, d0, v0):
        d0 = None if d0 is None else np.atleast_1d(d0)
        v0 = None if v0 is None else np.atleast_1d(v0)
        return d0, v0

    def _init_dva_part(self, nt, F0, d0, v0, static_ic, istime=True):
        if F0.shape[0] != self.n:
            raise ValueError(
                f"Initial force vector has {F0.shape[0]} elements;"
                f" {self.n} elements are expected"
            )
        if self.pre_eig:
            raise NotImplementedError(
                f"{type(self).__name__} generator not yet implemented "
                "using the `pre_eig` option"
            )

        d0, v0 = self._set_initial_cond(d0, v0)
        d, v, a = self._alloc_dva(nt, istime)
        f = np.copy(a)  # not a.copy because of `order` default
        f[:, 0] = F0
        self._init_dv(d, v, d0, v0, F0, static_ic)
        if self.rfsize:
            if self.unc:
                d[self.rf, 0] = self.ikrf.ravel() * F0[self.rf]
            else:
                d[self.rf, 0] = la.lu_solve(self.ikrf, F0[self.rf], check_finite=False)
        return d, v, a, f

    def _init_dva(
        self, force, d0, v0, static_ic, istime=True, freq=None, rf_disp_only=False
    ):
        if force.shape[0] != self.n:
            raise ValueError(
                f"Force matrix has {force.shape[0]} rows; {self.n} rows are expected"
            )

        d0, v0 = self._set_initial_cond(d0, v0)
        d, v, a = self._alloc_dva(force.shape[1], istime)

        if self.pre_eig:
            force = self.phi.T @ force

        self._init_dv(d, v, d0, v0, force[:, 0], static_ic)

        if self.rfsize:
            rf = self.rf
            if self.unc:
                d[rf] = self.ikrf * force[rf]
            else:
                d[rf] = la.lu_solve(self.ikrf, force[rf], check_finite=False)
            if not istime and not rf_disp_only:
                freqw = 2 * np.pi * freq
                freqw2 = freqw ** 2
                a[rf] = d[rf] * -(freqw2)
                v[rf] = d[rf] * (1j * freqw)

        return d, v, a, force

    def _calc_acce_kdof(self, d, v, a, force):
        """Calculate the `kdof` part of the acceleration"""
        if self.ksize:
            kdof = self.kdof
            F = force[kdof]
            if self.unc:
                if self.cdforces:
                    b = self.bo.copy()
                    i = np.arange(self.ksize)
                    b[i, i] = self.b  # full damping
                    B = b @ v[kdof]
                else:
                    B = self.b[:, None] * v[kdof]
                K = self.k[:, None] * d[kdof]
                if self.m is not None:
                    a[kdof] = self.invm * (F - B - K)
                else:
                    a[kdof] = F - B - K
            else:
                B = self.b @ v[kdof]
                K = self.k @ d[kdof]
                if self.m is not None:
                    a[kdof] = la.lu_solve(self.invm, F - B - K, check_finite=False)
                else:
                    a[kdof] = F - B - K

    def _add_rf_flex(self, flex, phi, velo, unc):
        if not velo and self.rfsize:
            rf = self.rf
            ikrf = self.ikrf
            phirf = phi[:, rf]
            if unc:
                flexrf = ikrf.ravel()[:, None] * phirf.T
            else:
                flexrf = la.lu_solve(ikrf, phirf.T, check_finite=False)
            flex = flex + phirf @ flexrf
        return flex

    def _flex(self, flex, phi):
        if isinstance(flex, float):
            n = phi.shape[0]
            flex = np.zeros((n, n))
        return flex

    def _force_freq_compat_chk(self, force, freq):
        """Check compatibility between force matrix and freq vector"""
        if force.shape[1] != len(freq):
            raise ValueError(
                f"Number of columns `force` ({force.shape[1]}) does "
                f"not equal length of `freq` ({len(freq)})"
            )
